package RFC::Alert;
#$Id: Alert.pm,v 1.11 2015/04/05 10:56:05 sqko Exp $
#################################################################
#
#
#       RFC::Alert
#
#
# The RFC::Alert is a device intended to be used for automatic 
# monitoring of process parameters.
#
# Each instance of RFC::Alert can monitor a single parameter (other
# RFC device) and if a given setpoint is passed, a single alert 
# mail is sent to recipients defined by either the alert_mail
# key for the RFC::Rig instance or the global system administrators
# if the alert_mail key is not defined.
#
# If an alert has occoured, the defvice can be set up to stop any
# running sequential progr4ams for the rig and/or execute a sequence
# of commands which potentially could rectify the situation leading
# up to the alert.
#
# The alert vill be canceled whenever the parameter beeing monitored
# passes the reset threshold.
#
# The alarm level and the reset level determines if the process 
# paratmeter have to ba aither above or below the alarm threshold for 
# an alarm ot trigger.
# If the reset level is below the alarm threshold, an alarm is triggered
# whenever the process paramater gets above the alarm threshold.
# If the alarm trheshold is below the reset level, the reverse is the case.
#
# RFC::Alarm instances are not included in normal data logging, as the 
# values of the monitored device will be logged by that device iteself.
#
# Complex trigger alarms are possible by combining several individual devices 
# through schmiodt trigger devices which themselves are connected by logic 
# devices, however care should be observed as the more complex the system 
# becomes the bigger the risk of misconfiguration resulting in loss of 
# experimental results (by premature shutdown of key devices for instance).
#
#################################################################
use strict;
use warnings;
use RFC::BaseDevice;
use RFC::Header;

our @ISA = qw(RFC::BaseDevice);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->die("second argument to new must be a RFC::Rig instance!") unless (ref($rig) eq 'RFC::Rig');
    $self->{'_type'} = 'Alert';
    $self->{'_setup_tags'} = ['device_type','device','title','threshold','reset','retries','kill_program','execute_commands','command_list'];
    $self->{'_setup_tags_readonly'} = [];
    $self->{'_mail'} = 1;
    $self->{'_desc'} = {
        'device_type' => 'Type of raw device',
        'device' => 'Raw device',
        'title' => 'Optional title',
	'threshold' => 'Threshold over (or under) which the devices triggeres an alert mail. If the threshold value is higher than the reset value the device value must be higher than the trigger level to trigger an alert, if threshold is lowet than reset value, a value below the threshold value triggers an alert.',
	'reset' => 'Value under (or over) the device resets the alert. If the threshold value is higher than the reset value the device value must be higher than the trigger level to trigger an alert, if threshold is lowet than reset value, a value below the threshold value triggers an alert.',
	'retries' => 'The number of retries performed before an alert is triggered (to awoid single read errors to triger alerts)',
	'kill_program' => 'Terminate any running sequential program if alert is triggered (only terminated once / alert)',
	'execute_commands' => 'Execute additional commands if alert is triggered (only executed once / alert)',
	'command_list' => 'List of commands to be executed upon alert trigger. Uses exactly same structure as normal sequential programs. Do NOT include comands which require a komma (,) in the argument list or any commands depending on external resources (timeslot and/or impedance commands etc)',
    };
    $self->{'_default'} = {
        'device_type' => 'simplechannel',
        'device' => '',
	'threshold' => 1,
	'reset' => 0.5,
	'retries' => 3,
	'kill_program' => 'No',
	'execute_commands' => 'No',
	'command_list' => '',
    };
    $self->{'_values_kill_program'} = ['Yes','No'];
    $self->{'_values_execute_commands'} = ['Yes','No'];
    bless $self,$class;
    my @list = ('');
    foreach my $t ($self->rig->list_types())
    {
	next unless ($t);
	### Exclude PID and Alert device classes as they make no sense using for input devices for a alert ###
	### For PID devices, anerts should be on the measure device directly and alerts of alert-devices   ###
	### is kind of pointless....                                                                       ###
 	next if ($t eq 'PID');
	next if ($t eq 'Alert');
	push @list,$t;
    }
    $self->{'_values_device_type'} = [@list];
    $self->{'_section'} = 'Alert_'.$self->name;
    $self->load_conf();
    $self->{'_values_device'} = ['',$self->rig->avaliable_devices($self->{'_device_type'})];
    if ($self->{'_device_type'} && $self->{'_device'})
    {
        my $dev = $self->rig->get_device($self->{'_device_type'},$self->{'_device'});
#	$self->die('Could not get or initialise device "'.$self->{'_device'}.'" of type "'.$self->{'_device_type'}.'"') unless ($dev);
        $self->{'_dev'} = $dev if ($dev);       
    }
    my $recipients = $self->rig->get_cv('main','warning_mails');
    if ($recipients)
    {
	$self->{'_info_string'} = 'Alert mail will be sent to: '.$recipients;
    }
    else
    {
	$self->{'_info_string'} = 'Warning, no recipients defined for warning mail! Alerts will only be sent to system administrators! (please setup recipients under miscelaneous setup)';
    }
    return $self;
}

sub check
{
    my $self = shift;
    return undef unless ($self->device());
    my $val = $self->read(@_);
    my $count = $self->{'_retries'};
    if ($self->{'_threshold'} < $self->{'_reset'})
    {
	if ($val > $self->{'_threshold'})
	{
	    $self->rig->unset_warning($self->{'_section'});
	    return 0;
	}
	return 1 if $self->rig->check_warning($self->{'_section'});
	while ($count--)
	{
	    $val = $self->read();
	    return 0 if ($val > $self->{'_threshold'});
	} 
	return $self->alert("Value for '".$self->device->name."' is below threshold (".$self->{'_threshold'}.").\nMeasured value: $val",'Alert from rig '.$self->rig->name.", Alert device '".$self->name. "' has triggered!");
    }
    else
    {
	if ($val < $self->{'_threshold'})
	{
	    $self->rig->unset_warning($self->{'_section'});
	    return 0;
	}
	return 1 if $self->rig->check_warning($self->{'_section'});
	while ($count--)
	{
	    $val = $self->read();
	    return 0 if ($val < $self->{'_threshold'});
	}
	return $self->alert("Value for '".$self->device->name."' is above threshold (".$self->{'_threshold'}.").\nMeasured value: $val",'Alert from rig '.$self->rig->name.", Alert device '".$self->name. "' has triggered!");
    }
}


sub test
{
    my $self = shift;
    return $self->device->test(@_);
}


sub test_string
{
    my $self = shift;
    $self->{'_mail'} = 0;
    my $str = $self->check();
    $str = 'Status for alert device '.$self->name.': Ok. Device value: '.$self->device->read() unless ($str); 
    if ($str == 1)
    {
	$str = 'Status for alert device '.$self->name.': Not ok. Alert mail(s) has been sent. Device value: '.$self->device->read(); 
    }
    $self->{'_mail'} = 1;
    return $str;
}


sub kill_program()
{
    my $self = shift;
    return 0 unless ($self->{'_kill_program'} && $self->{'_kill_program'} eq 'Yes');
    my $cmd = '/bin/ps -efl | /bin/grep custom_prog.pl';
    my @list = `$cmd`;
    $self->writedebug(@list) if ($self->debug);
    my $r = $self->rig->name();
    for my $st (@list)
    {
        chomp $st;
        ### Finding the right processid to kill ###
        if ($st =~ /^\d+\s+\w+\s+rig$r\s+(\d+)\s+.*\/custom_prog\.pl\s+$r\s+/ )
        {
            my $pid = $1+0;
            ### Using a SIGINT (kill -s 2) to emulate a Ctrl+C to custom_prog.pl ###
            system("/bin/kill -s 2 $pid &");
	    return 1;
        }   
    }
    return 0;
}

sub execute_cmds()
{
    my $self = shift;
    return 0 unless ($self->{'_execute_commands'} && $self->{'_execute_commands'} eq 'Yes');
    return 0 unless ($self->{'_command_list'});
    my @lines = split(/\,/,$self->{'_command_list'});
    my $file = '/tmp/Rig'.$self->rig->name.'_Alert_'.$self->name;
    open ALERT,'>',$file or $self->die("could not open file $file for writing $!");
    print ALERT join("\n",@lines);
    close ALERT;
    my $cmd = $RFC::Header::libdir.'custom_prog.pl '.$self->rig->name.' '.$file.' >/dev/null 2>>/dev/null &';
    system($cmd);
    return 1;
}

sub alert
{
    my $self = shift;
    my $message = shift;
    my $subj = shift;
    if ($self->{'_mail'})
    {
	$message .= "\nRunning program terminated!" if ($self->kill_program());
	$message .= "\nTrigger commands activated!" if ($self->execute_cmds());
	$self->rig->warning_mail($self->{'_section'},$message,$subj);
	return 1;
    }
    else
    {
	return $subj."\n".$message;
    }
}

sub device()
{
    my $self = shift;
    return $self->{'_dev'};
}


sub read
{
    my $self = shift;
    return $self->device->read(@_) if ($self->device());
    return -32768;
}
