package RFC::AnalogOut;
#$Id: AnalogOut.pm,v 1.45 2015/03/02 11:54:19 sqko Exp $
##########################################################################################
#
#
#   RFC::AnalogOut
#
#
# This class defines simple analog output devices (D-A converters, not power supplies)
# To obtain a analog output instance call onw of the constructors:
#
# $id = RFC::AnalogOut->new($name,$rig);
#
# $id = RFC::AnalogICP87024->new($name,$rig);
#
# $id = RFC::AnalogICP87028->new($name,$rig);
#
# $id = RFC::ManualAnalogOut->new($name,$rig); (A virtual, software only, device)
#
# The rig argument must be a RFC::Rig instance or a similar object honoring
# the get_conf() member function.
# The constructiors can also be called with additional arguments like shown below:
#  
# $id = RFC::AnalogICP87024->new($name,$rig,$tty,$address,$channel);
#
# where the last 3 argmeutns specify the serial device, the module address 
# and the channel number of the output module (as most modules contains more 
# than one output channel).
#
#  All analog output objects have the following member functions (excluding those 
#  derived from RFC::BaseDevce and Debug.pm):
#
#  $id->set($)    Sets the output voltage/current of the device to the specified argument.
#                 the format for sending numbers to the devices supported (i87022/24/26)
#                 is the engineering unit DD.DDD according to the manual. The driver
#                 uses sprintf to make sure that the supplied number is in this format.
#
#
###########################################################################################

use strict;
use warnings;
use RFC::BaseDevice;
use RFC::Cache;
our @ISA = qw(RFC::BaseDevice);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_type'} = 'AnalogOut';
    $self->{'_mode'} = 'Automatic';
    $self->{'_values_mode'} = ['Volt','Amp'];
    $self->{'_setup_tags'} = ['mode','tty','address','channel','title','show_plot','use_cache'];
    $self->{'_desc'} = {
        'mode' => 'Device control mode',
        'tty' => 'Serial device for communication (ex ttyS0)',
        'address' => 'Device address',
        'channel' => 'Device channel',
	'title' => 'Optional title',
        'show_plot'=>'Determines if the current device data is to be shown in the daily data plots',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
        'mode' => 'Automatic',
        'tty' => 'ttyS0',
        'address' => 0,
        'channel' => 0,
        'show_plot'=>'Yes',
        'use_cache'=>'No',
    };
    $self->{'_set_fn_names'} = ['set']; # List of function names which can trigger a notify event
    $self->{'_values_mode'} = ['Automatic'];
    $self->{'_max'} = 0;
    $self->{'_min'} = 0;
    bless $self,$class;
    foreach my $key ('tty','address','channel','title','show_plot','use_cache','title')
    {
        $self->{'_'.$key} = $self->get_conf($key);
        $self->{'_'.$key} = $self->get_default($key) unless ($self->{'_'.$key});
    }
    $self->{'_tty'} = 'ttyS0' unless ($self->{'_tty'});
    $self->{'_address'} = '0' unless ($self->{'_address'});
    $self->{'_channel'} = '0' unless ($self->{'_channel'});
    $self->{'_section'} = 'analog_out_'.$self->name;
    return $self;
}

sub test_string()
{
    my $self = shift;
    $self->init unless ($self->mode !~ /specified/);
    my $value = $self->read(@_);
    return 'Analog out '.$self->name.": $value ".$self->mode."\n";
}

sub set($)
{
    my $self = shift;
    my $value = shift;
    $self->init unless ($self->mode);
    $self->writedebug("Setting analog output for ".$self->name." to $value") if ($self->debug > 2);
    $self->die($self->{'_type'}." is a base class and does not impelemnt id->set(), only the derived classes does so!");
    return undef;
}

sub setup_mode()
{
    my $self = shift;
    $self->init unless ($self->mode !~ /specified/);
    return 'Output: '.$self->{'_min'}.' to '.$self->{'_max'}.' '.$self->mode();
}

sub print_config()
{
    my $self = shift;
    my @arr = ('Analog output type: '.$self->type);
    push @arr,'Name: '.$self->name;
    push @arr,'TTY: '.$self->{'_tty'};
    push @arr,'Address: '.$self->{'_address'};
    push @arr,'channel: '.$self->{'_channel'};
    push @arr, $self->setup_mode;
    return join (', ',@arr);
}

sub init()
{
    my $self = shift;
}

package RFC::ManualAnalogOut;
use RFC::Header;
our @ISA = qw(RFC::AnalogOut);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    bless $self,$class;
    $self->{'_type'} = 'ManualAnalogOut';
    $self->{'_mode'} = 'Volt/mA';
    $self->{'_setup_tags'} = ['mode','title','show_plot','use_cache'];
    $self->{'_desc'} = {
        'mode' => 'Device control mode',
	'title' => 'Optional title',
        'show_plot'=>'Determines if the current device data is to be shown in the daily data plots',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
        'mode' => 'Volt/mA',
        'show_plot'=>'Yes',
        'use_cache'=>'No',
    };
    $self->{'_values_mode'} = ['Volt/mA'];
    foreach my $key ('mode','title','show_plot','use_cache')
    {
        $self->{'_'.$key} = $self->get_conf($key);
        $self->{'_'.$key} = $self->get_default($key) unless ($self->{'_'.$key});
    }
    $self->{'_type'} = 'ManualAnalogOut';
    $self->{'_file'} = SemaforeFile->new($userdir.'rig'.$self->rig->name.'/analog_'.$self->name,'/home/celltest/Gas_rig'.$self->rig->name.'.lock');
    unless ($self->{'_file'}->exist)
    {
        ### Create file with rw for all           ###
        ### Yes I know 666 is the number of evil! ###
        $self->{'_file'}->lock_ex;
        $self->{'_file'}->writeline('0');
        chmod 0666, $self->{'_file'}->filename;
        $self->{'_file'}->unlock;
    }
    return $self;
}

sub DESTROY
{
    my $self = shift;
    $self->{'_file'}->DESTROY() if ($self->{'_file'});
    return undef;
}


sub test()
{
    return 1;
}

sub set($)
{
    my $self = shift;
    my $status = shift;
    $status = $self->set_noinfo($status);
    if (@_)
    {
        my $user = shift;
        $self->rig->proglog("Analog output ".$self->name." set to $status manually",$user) unless (@_);
    }
    else
    {
        $self->rig->proglog("Analog output ".$self->name." set to $status manually");
    }
    $self->notify('set',$status);
    return $status;
}

sub set_noinfo($)
{
    my $self = shift;
    my $status = shift;
    $self->writedebug("Setting status for relay ".$self->name." to $status manually!") if ($self->debug > 2);
    $self->{'_file'}->writeline($status);
    $self->notify('set',$status);
    return $status;
}

sub read_nocache
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	my $data = RFC::Cache->new();
	$data->debug($self->debug()) if ($self->debug);
    }
    my $flow = 0;
    $self->check_cache($data);
    my $cached = $data->get('manual_Analogout',$self->name);
    return $cached if (defined($cached));
    my @lines = $self->{'_file'}->readlines;
    $flow = $lines[0] if (@lines);
    $data->set('manual_Analogout',$self->name,$flow);
    $self->set_cache($flow);
    return $flow;
}

sub print_config()
{
    my $self = shift;
    my @arr = ('Analog output type: '.$self->type);
    push @arr,'Name: '.$self->name;
    push @arr,'Mode: '.$self->mode;
    return join (', ',@arr);
}


package RFC::ICPAnalogOut;
our @ISA = qw(RFC::AnalogOut);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    bless $self,$class;
    foreach my $key ('tty','address','channel','mode','title','show_plot','use_cache')
    {
        $self->{'_'.$key} = $self->get_conf($key);
        $self->{'_'.$key} = $self->get_default($key) unless ($self->{'_'.$key});
    }
    $self->{'_type'} = 'ICPAnalogOut';
    $self->{'_tty'} = shift if (@_);
    $self->{'_address'} = shift if (@_);
    $self->{'_channel'} = shift if (@_);
    return $self;
}

sub test()
{
    my $self = shift;
    $self->init unless ($self->mode !~ /specified/);
    return 0 unless ($self->{'_min'} != 0 || $self->{'_max'} != 0);
    return 1;
}

sub set($)
{
    my $self = shift;
    my $out = shift;
    $self->init unless ($self->mode !~ /specified/);
    if ($out < $self->{'_min'})
    {
	$self->writedebug("Output below minimum value ($self->{'_min'}), setting minimum");
	$out = $self->{'_min'};
    }
    if ($out > $self->{'_max'})
    {
	$self->writedebug("Output above maximum value ($self->{'_mmax'}), setting maximum");
	$out = $self->{'_min'};
    }
    $self->writedebug("Setting output for ".$self->name." to $out") if ($self->debug > 2);
    my $res = SocketClient::serial_client_raw($self->{'_tty'},'volt_set',$self->{'_address'},$self->{'_channel'},sprintf("%06.3f",$out));
    $self->writedebug("Response for relay ".$self->name.": $res") if ($self->debug > 3);
    $self->notify('set',$out);
    return $res;
}

sub read_nocache
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	my $data = RFC::Cache->new();
	$data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    my $ch = $self->{'_channel'};
    my $cached = $data->get($tty,'icp',$addr,$ch);
    if (defined($cached))
    {
        $self->writedebug("Using existing data from channel $ch output on tty $tty address $addr") if ($self->debug > 3);
        $res = $cached;
    }
    else
    {
        $self->writedebug("Reading data from relay on tty $tty address $addr") if ($self->debug > 3);
	my $cmd = sprintf("\$%02d6%d",$self->{'_address'},$self->{'_channel'});
        $res = SocketClient::serial_client_raw($self->{'_tty'},'icp_raw',$cmd);
	chomp $res;
	if ($res =~ /([\+\-][\d\.]+)/)
	{
	    $res = $1+0;
	}
        if (defined($data)) 
        {
	    $data->set($tty,'ICP',$addr,$ch,$res);
        }
    }
    $self->die("No data returned from device!") unless (defined($res));
    $self->writedebug("Value for analog output ".$self->name." was ".$res) if ($self->debug > 2);
    my $str = time .' '. $self->name .' '.$res;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    $self->set_cache($res);
    return $res;
}

sub read_active()
{
    my $self = shift;
    my $cmd = sprintf("\$%02d6%d",$self->{'_address'},$self->{'_channel'});
    my $res = SocketClient::serial_client_raw($self->{'_tty'},'icp_raw',$cmd);
    chomp $res;
    if ($res =~ /([\+\-][\d\.]+)/)
    {
	return $1+0;
    }
    $self->die("No data returned from device!") unless (defined($res));
    my $str = time .' '. $self->name .'_active '.$res;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    retrun $res;
}

sub config()
{
    my $self = shift;
    my $cmd = sprintf("\$%02d2",$self->{'_address'});
    my $res = SocketClient::serial_client_raw($self->{'_tty'},'icp_raw',$cmd);
    ### Try 2 times to get configuration ###
    unless (defined($res) && $res =~ /\w/)
    {
	select(undef,undef,undef,0.5);
	$res = SocketClient::serial_client_raw($self->{'_tty'},'icp_raw',$cmd) 
    }
    chomp $res if (defined($res));
    return $res;
}

package RFC::AnalogICP87024;
our @ISA = qw(RFC::ICPAnalogOut);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_tty'} = shift if (@_);
    $self->{'_address'} = shift if (@_);
    $self->{'_channel'} = shift if (@_);
    $self->{'_type'} = 'AnalogICP87024';
    bless $self,$class;
    if (@_)
    {
	$self->{'_min'} = shift;	
	$self->{'_max'} = shift;
	$self->{'_mode'} = shift;
    }
    else
    {
	$self->{'_mode'} = 'Not specified';
    }
    return $self;
}

sub init()
{
    my $self = shift;
    my $str = $self->config;
    if (defined($str) && $str =~ /\!\d{2}3(\d)/)
    {
	my $c = $1;
	if ($c == 0)
	{
	    $self->{'_min'} = 0;
	    $self->{'_max'} = 20;
	    $self->{'_mode'} = 'mA';
	}
	if ($c == 1)
	{
	    $self->{'_min'} = 4;
	    $self->{'_max'} = 20;
	    $self->{'_mode'} = 'mA';
	}
	if ($c == 2)
	{
	    $self->{'_min'} = 0;
	    $self->{'_max'} = 10;
	    $self->{'_mode'} = 'Volt';
	}
	if ($c == 3)
	{
	    $self->{'_min'} = -10;
	    $self->{'_max'} = 10;
	    $self->{'_mode'} = 'Volt';
	}
	if ($c == 4)
	{
	    $self->{'_min'} = 0;
	    $self->{'_max'} = 5;
	    $self->{'_mode'} = 'Volt';
	}
	if ($c == 5)
	{
	    $self->{'_min'} = -5;
	    $self->{'_max'} = 5;
	    $self->{'_mode'} = 'Volt';
	}
    }
    else
    {
	$str = 'UNDEF' unless defined($str);
	$self->writedebug("Init failed for device ".$self->name.": returned '$str'");
    }
    return $self;
}

package RFC::AnalogICP87028;
our @ISA = qw(RFC::ICPAnalogOut);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_tty'} = shift if (@_);
    $self->{'_address'} = shift if (@_);
    $self->{'_channel'} = shift if (@_);
    $self->{'_type'} = 'AnalogICP87028';
    bless $self,$class;
    if (@_)
    {
	$self->{'_min'} = shift;	
	$self->{'_max'} = shift;
	$self->{'_mode'} = shift;
    }
    else
    {
	$self->{'_mode'} = 'Not specified';
    }
    return $self;
}

sub init()
{
    my $self = shift;
    my $str = $self->config;
    if (defined($str) && $str =~ /\!\d{2}3(\w)/)
    {
	my $c = $1;
	if ($c eq "F")
	{
	    $self->{'_min'} = 0;
	    $self->{'_max'} = 20;
	    $self->{'_mode'} = 'mA';
	}
    }
    else
    {
	$str = 'UNDEF' unless defined($str);
	$self->writedebug("Init failed for device ".$self->name.": returned '$str'");
    }
    return $self;
}

return 1;


