package RFC::Logic;
#$Id: Logic.pm,v 1.21 2015/03/05 07:08:35 sqko Exp $
#################################################################
#
#
#       RFC::Logic
#
#
# The RFC::Logic module defines a list of virtual logical devices
# Each of the devices implements one of the logical operators
# AND, OR, XOR etc.
#
# A logic device operates on logical inputs, which can be logical
# devices, relay devices or the special filter device 'Schmidt trigger'
# which converts a floating point value to a logical value based
# on specific threshold values.
#
# A RFC::Logic derived device can have from 1 to 9 inputs and an 
# optional single output device. The output device must be a relay 
# device type and is intended to convey the result of one or more
# logical operations to the actual control system. 
#
# As RFC::Logic devices are virtual, they are not included in the 
# normal datalogging (and accordingly readstring_ignore(1) is set
# uppon device instantiation). However they must be enabled
# in the rig configuration to work properly.
#
# Interdevice control flow is implemented by the obsever pattern.
# This ensures that input devices does not need to know that they 
# are used as inputs to logic devices ad any set() or set_noinfo()
# command exeuted on a relay device automatically forces ay logic
# device using this as an input to get updated (and potentially
# forcing updates on logic devices further 'down the chain' to be
# updated).
#
# Notice however that great care must be taken when including logic 
# devices as it is easy to configure a situation where circular 
# references will occour (essentially causing Perl's version of
# a stack overflow).
#
# Tus it is recommended that if logic devices is to be included in 
# a rig's control system, make a detailed schematic of the control
# flow diagram with unique names of each individual logic gate
# thus ensuring that the correct inputs and gate names are chosen
# when configuring the rig.
#
#################################################################


use strict;
use warnings;
use RFC::Header;
use RFC::BaseDevice;
our @ISA = qw(RFC::BaseDevice);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_type'} = 'Generic logic device'; 
    $self->{'_setup_tags'} = ['class','mode','inputs','output_device'];
    $self->{'_setup_tags_readonly'} = ['mode'];
    $self->{'_desc'} = {
        'mode' => 'Device control mode',
	'inputs' => 'number of inputs',
        'output_device' => 'Relay device for output (optional if current device is used in a logic chain)',
    };
    $self->{'_default'} = {
	'mode' => 'Logic',
	'inputs' => 2,
    };
    bless $self,$class;
    $self->{'_values_class'} = [RFC::Device::list_Logic()];
    $self->{'_values_inputs'} = [2];
    $self->{'_values_mode'} = ['Logic'];
    $self->{'_device_type'} = 'relay';
    $self->{'_section'} = 'Logic_'.$self->name;
    foreach my $tag ($self->setup_tags)
    {
        $self->{'_'.$tag} = $self->get_default($tag);
        my $conf = $self->get_conf($tag);
        $self->{'_'.$tag} = $conf if (defined($conf) && $conf =~ /\w/);
    }
    my @list = ();
    ### Load list of relays ###
    foreach ($self->rig->avaliable_devices($self->{'_device_type'}))
    {
	next unless ($_);
        push @list,$_ unless ($_ eq $self->name()); ## Do not include self ##
    }
    $self->{'_values_output_device'} = ['',@list];
    if ($self->{'_device_type'} && $self->{'_output_device'})
    {
        ### get device if it already exist ###
        my $dev = $self->rig->get_device($self->{'_device_type'},$self->{'_output_device'});
	if ($dev)
	{
	    $self->{'_dev'} = $dev;
	    $dev->set_readonly();
	}
    }
    ### Include any Logic devices as possible inputs ###
    foreach ($self->rig->avaliable_devices('logic'))
    {
	next unless ($_);
        push @list,$_ unless ($_ eq $self->name()); ## Do not include self ##
    }
    ### Include any Schmidt trigger devices as possible inputs ###
    foreach my $n ($self->rig->avaliable_devices('filter'))
    {
	next unless ($n);
	my $d = $self->rig->init_device('filter',$n);
	next unless ($d);
	next unless ($d->type eq 'Schmidt trigger latch');
        push @list,$n unless ($n eq $self->name()); ## Do not include self ##
    }
    ### Remove any duplicates in list (may happen as logic devices register as relays) ###
    my %hash = ();
    foreach (@list)
    {
	$hash{$_} = 1;
    }
    @list = ();
    foreach my $key (sort keys %hash)
    {
	push @list,$key;
    }
    for (my $x = 0; $x < $self->{'_inputs'}; $x++)
    {
	my $tag = 'input_device_'.$x;
	push @{$self->{'_setup_tags'}},$tag;
        $self->{'_'.$tag} = $self->get_default($tag);
        my $conf = $self->get_conf($tag);
        $self->{'_'.$tag} = $conf if (defined($conf) && $conf =~ /\w/);
	$self->{'_desc'}->{$tag} = 'Input device '.$x;
	$self->{'_values_'.$tag} = ['',@list];
    }
    $self->rig->register_device('relay',$self);
    ## Note logic devices should never be logged ###
    $self->readstring_ignore(1);
    $self->{'_readonly'} = 1;
    $self->{'_info_string'} = 'Note that logic devices only works correctly if they are included (enabled)';
    ### Note, do NOT run init as part of constructor! ###
    return $self;
}

sub setup_tags()
{
    my $self = shift;
    my @res = ();
    @res = @{$self->{'_setup_tags'}} if ($self->{'_setup_tags'});
    return (@res,'comments');
}



sub init()
{
    my $self = shift;
    for (my $x = 0; $x < $self->{'_inputs'}; $x++)
    {
	next if ($self->{'_dev_'.$x}); 
	my $tag = '_input_device_'.$x;
	my $dev = $self->rig->get_device($self->{'_device_type'},$self->{$tag});
	$dev = $self->rig->get_device('logic',$self->{$tag}) unless ($dev);
	$dev = $self->rig->get_device('filter',$self->{$tag}) unless ($dev);
	unless ($dev)
	{
	    $dev = $self->rig->init_device($self->{'_device_type'},$self->{$tag});
	    $dev = $self->rig->init_device('logic',$self->{$tag}) unless ($dev);
	    $dev = $self->rig->init_device('filter',$self->{$tag}) unless ($dev);
	    $dev->readstring_ignore(1) if ($dev);
	}
	if ($dev)
	{
	    $dev->debug($self->debug);
	    $self->{'_dev_'.$x} = $dev; 
	    $dev->attatch_observer(\$self);
	}
	$self->die("Could not initialise Logic device of type '".$self->{'_device_type'}."' with name: '".$self->{'_device'}."'") unless ($self->{'_dev_'.$x});
    }
    unless ($self->{'_dev'})
    {
	my $tag = '_output_device';
	my $dev = $self->rig->get_device($self->{'_device_type'},$self->{$tag});
	unless ($dev)
	{
	    $dev = $self->rig->init_device($self->{'_device_type'},$self->{$tag});
	    $dev->readstring_ignore(1) if ($dev);
	}
	if ($dev)
	{
	    $dev->debug($self->debug);
	    $self->{'_dev'} = $dev; 
	    $dev->set_readonly();
	}
    }
    return $self;
}

sub print_config()
{
    my $self = shift;
    $self->init();
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Type: '.$self->type;
    for (my $x = 0; $x < $self->{'_inputs'}; $x++)
    {
	my $tag = '_dev_'.$x;
	push @arr,'Input device '.$x.': '.$self->{$tag}->print_config() if ($self->{$tag});
    }
    push @arr,'Output device: '.$self->{'_dev'}->print_config() if ($self->{'_dev'});
    return join(', ',@arr);
}


sub local_init()
{
    #### Stub function intended for loverloading for any impelmention dependent initialisations ###
    my $self = shift;
    return $self;
}

sub debug
{
    my $self = shift;
    $self->SUPER::debug(@_);
    if (@_)
    {
	for (my $x = 0; $x < $self->{'_inputs'}; $x++)
	{
	    my $tag = '_dev_'.$x;
	    $self->{$tag}->debug(@_) if ($self->{$tag});
	}
    }
    $self->{'_dev'}->debug(@_) if ($self->{'_dev'});
    return $self->{'_debug'};
}

sub readonly_value($)
{
    my $self = shift;
    my $tag = shift;
    return undef unless ($tag);
    ### allow call to get 'mode' for undefined devices ###
    return 'Automatic' if ($tag eq 'mode');
    return 'relay' if ($tag eq 'device_type');
}

sub observe($$)
{
    my $self = shift;
    my $action = shift; ## Must always be 'set' or 'set_noinfo' as logic devices emulates relays;
    my $value = shift;
    $self->set($value) if ($action eq 'set');
    $self->set_noinfo($value) if ($action eq 'set_noinfo');
}


sub set($)
{
    my $self = shift;
    my $res = shift;
    my $val = $self->read();
    $self->notify('set',$val);
    $self->{'_dev'}->set($val) if ($self->{'_dev'});
    return $res;
}

sub set_noinfo($)
{
    my $self = shift;
    my $res = shift;
    my $val = $self->read();
    $self->notify('set_noinfo',$val);
    $self->{'_dev'}->set_noinfo($val) if ($self->{'_dev'});
    return $res;
}

sub read()
{
    my $self = shift;
    $self->init();
    ### The default Logic device always returns 1 on read.... ###
    return 1;
}

sub test()
{
    my $self = shift;
    $self->init();
    my $res = 1;
    for (my $x = 0; $x < $self->{'_inputs'}; $x++)
    {
	my $tag = '_dev_'.$x;
	$res = 0 unless ($self->{$tag}->test);
    }
    return $res;
}

sub test_string()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug) if ($self->debug);
    }
    $self->check_cache($data);
    $self->init(); 
    my @arr = ();
    my $type = $self->type();
    my $string = 'Logic device "'.$self->name.'" of type "'.$type.'": ';
    my $val = $self->read($data);
    push @arr,'Value: '.$val;
    for (my $x = 0; $x < $self->{'_inputs'}; $x++)
    {
	my $tag = '_dev_'.$x;
	my $str = $self->{$tag}->test_string($data);
	chomp $str;
	push @arr,'Input device '.$x.': '.$str;
    }
    push @arr,'Output device: '.$self->{'_dev'}->test_string($data) if ($self->{'_dev'});
    $string .= join(', ',@arr)."\n";
    return $string;
}

package RFC::Logic_AND;
our @ISA = qw(RFC::Logic);


sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'AND logic device';
    $self->{'_default'}->{'class'} = 'AND';
    $self->{'_values_inputs'} = [2,3,4,5,6,7,8,9];
    return $self;
}

sub read
{
    my $self = shift;
    $self->init();
    my $res = 1;
    for (my $x = 0; $x < $self->{'_inputs'}; $x++)
    {
	my $tag = '_dev_'.$x;
	unless ($self->{$tag}->read(@_))
	{
	    $res = 0;
	    last;
	}
    }
    return $res;
}

package RFC::Logic_NAND;
our @ISA = qw(RFC::Logic_AND);


sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'NAND logic device';
    $self->{'_default'}->{'class'} = 'NAND';
    return $self;
}

sub read
{
    my $self = shift;
    my $val = $self->SUPER::read(@_);
    if ($val)
    {
	$val = 0;
    }
    else
    {
	$val = 1;
    }
    return $val;
}



package RFC::Logic_OR;
our @ISA = qw(RFC::Logic);


sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'OR logic device';
    $self->{'_default'}->{'class'} = 'OR';
    $self->{'_values_inputs'} = [2,3,4,5,6,7,8,9];
    return $self;
}

sub read
{
    my $self = shift;
    $self->init();
    my $res = 0;
    for (my $x = 0; $x < $self->{'_inputs'}; $x++)
    {
	my $tag = '_dev_'.$x;
	if ($self->{$tag}->read(@_))
	{
	    $res = 1;
	    last;
	}
    }
    return $res;
}

package RFC::Logic_NOR;
our @ISA = qw(RFC::Logic_OR);


sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'NOR logic device';
    $self->{'_default'}->{'class'} = 'NOR';
    return $self;
}

sub read
{
    my $self = shift;
    my $val = $self->SUPER::read(@_);
    if ($val)
    {
	$val = 0;
    }
    else
    {
	$val = 1;
    }
    return $val;
}


package RFC::Logic_XOR;
our @ISA = qw(RFC::Logic);


sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'XOR logic device';
    $self->{'_default'}->{'class'} = 'XOR';
    $self->{'_values_inputs'} = [2];
    return $self;
}

sub read
{
    my $self = shift;
    $self->init();
    my $val0 = $self->{'_dev_0'}->read;
    my $val1 = $self->{'_dev_1'}->read;
    my $res = 1;
    $res = 0 if ($val0 == $val1);
    return $res;
}

package RFC::Logic_NXOR;
our @ISA = qw(RFC::Logic_XOR);


sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'NXOR logic device';
    $self->{'_default'}->{'class'} = 'NXOR';
    return $self;
}

sub read
{
    my $self = shift;
    my $val = $self->SUPER::read(@_);
    if ($val)
    {
	$val = 0;
    }
    else
    {
	$val = 1;
    }
    return $val;
}


package RFC::Logic_NOT;
our @ISA = qw(RFC::Logic);


sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    bless $self,$class;
    $self->{'_type'} = 'NOT logic device';
    $self->{'_setup_tags'} = ['class','mode','device_type','inputs','input_device_0','output_device'];
    $self->{'_default'} = {
	'mode' => 'Logic',
	'inputs' => 1,
        'device_type' => 'relay',
	'class' => 'NOT',
    };
    $self->{'_values_inputs'} = [1];
    foreach my $tag ($self->setup_tags)
    {
        $self->{'_'.$tag} = $self->get_default($tag);
        my $conf = $self->get_conf($tag);
        $self->{'_'.$tag} = $conf if (defined($conf) && $conf =~ /\w/);
    }
    my @list = ();
    foreach ($self->rig->avaliable_devices($self->{'_device_type'}))
    {
	next unless ($_);
        push @list,$_ unless ($_ eq $self->name()); ## Do not include self ##
    }
    ### Include any Logic devices as possible inputs ###
    foreach ($self->rig->avaliable_devices('logic'))
    {
	next unless ($_);
        push @list,$_ unless ($_ eq $self->name()); ## Do not include self ##
    }
    ### Include any Schmidt trigger devices as possible inputs ###
    foreach my $n ($self->rig->avaliable_devices('filter'))
    {
	next unless ($n);
	my $d = $self->rig->init_device('filter',$n);
	next unless ($d);
	next unless ($d->type eq 'Schmidt trigger latch');
        push @list,$n unless ($n eq $self->name()); ## Do not include self ##
    }
    ### Remove any duplicates in list (may happen as logic devices register as relays) ###
    my %hash = ();
    foreach (@list)
    {
	$hash{$_} = 1;
    }
    @list = ();
    foreach my $key (sort keys %hash)
    {
	push @list,$key;
    }
    my $tag = 'input_device_0';
    $self->{'_'.$tag} = $self->get_default($tag);
    my $conf = $self->get_conf($tag);
    $self->{'_'.$tag} = $conf if (defined($conf) && $conf =~ /\w/);
    $self->{'_desc'}->{$tag} = 'Input device 0';
    $self->{'_values_'.$tag} = ['',@list];
    delete($self->{'_dev_1'});
    return $self;
}

sub read
{
    my $self = shift;
    $self->init();
    my $val0 = $self->{'_dev_0'}->read;
    my $res = 1;
    $res = 0 if ($val0);
    return $res;
}

return 1;
