package RFC::Rig;
#$Id: Rig.pm,v 1.237 2015/03/30 10:01:32 sqko Exp $

###################################################################
#
#
#
#     Package RFC::Rig;
#
#
#
# This class manages the complete setup ofr a RFCcontroll rig
# To obtain an instance call the constructor:
#
# $id = RFC::Rig->new($rignr)
#
# Where $rignr is an integer.
#
# -- \subsection{Public member functions}
#
# Each Rig instance has the following public member functions:
#
# $id->lock_control()         Locks the semafore file asociated with 
#                             the rig instance control.
# $id->unlock_control()       Unlocks the semafore for rig control.
# $id->die($)                 Wraper for CORE::die, but dumps instance 
#                             setup.
# $id->config()               Returns the configuration object of the 
#                             current test (An ElchemeaConfig instance).
# $id->begin()                Initiates a transaction.
# $id->commit()               Commits a transaction.
# $id->rollback()             Roll back a transaction.
# $id->error()                Returns the errorstring (if any).
# $id->load()                 Load the stored data (from file).
# $id->store()                Stores aquired data to file (Note, this 
#                             is not a part of the data logging!).
# $id->get_temp_kanaldata($)  Returns the time and data (as an array) 
#                             asociated with the specified key (returns 
#                             (undef,undef) if no data is found for that key).
# $id->riglock()              Returns the SemaforeFile object for the rig 
#                             instance
# $id->set_warning($)         Sets the specified string in the mailwarinig 
#                             file used to make sure that only one errormail
#                             is sent for each warning condition.
# $id->check_warning($)       Checks if the specified type is already in the
#                             warning mail file.
# $id->unset_warning($)       Removes the specified type from the warningmail
#                             file
# $id->is_iv()                Returns true if the iv_in_progress file exists
# $id->iv_start()             Retruns 1 (true) if the iv_in_progress file does
#                             not exist, and creates it in the process.
# $id->iv_stop()              Removes the iv_in_progress file (effectively stopping
#                             any programs / iv's locked to the existence of 
#                             this file).
# $id->iv_in_progress()       Returns the SemaforeFile object representing the
#                             iv_in_progress file. Can be used for using the 
#                             file for storing and retrieving process parameters.
# $id->warning_mail($type,$s,[$subj])  Sends a mail with content $s of type $type to 
#                             the recipients defined in the 'warning_mail' key
#                             in the 'main' section of the rig configuration.
#                             If no recipients are found the list defined in
#                             the 'errormails' key in the 'global' section of the 
#                             global configuration is used instead.
#                             Note that the mail is only sent if the $type
#                             is nor already defined in the warningmail file.
#                             If a third argument is supplied, this becomes the subject 
#                             of the mail, otherwise a default subject is used.
# $id->kanaldata()            Sets and gets the logged data (in conjnction with 
#                             data logging. If arguments are passed, they are
#                             assumed to be in the form returned by the readstring
#                             member function of a BaseDevice class instance.
# $id->readstring([$])        Does a complete data logging by callng the readstring
#                             member function on all asociated BaseDevice instances
#                             (This includes derived classes). This function handles 
#                             gracefully if a device instance dies while measuring,
#                             in which case an error is logged and a dummy value is
#                             returned for that device instead of the real value.
#                             This is to ensure that datalogging will proceed as 
#                             much as possible.
#                             If an argument is passed to readstring, it must be an 
#                             instance of RFC::Cache.
#                             Notice, that readstring does NOT write any data to file,
#                             it only returns a text string containing all the data.
#                             (Saving data to file can be handled by the normal Perl file
#                             handling functions).
# $id->list_sections()        Returns a list of configuration sections (names).
#                             If a argument is specified, only section names matching
#                             the specified string are returned.
# $id->config_section_exists($)  
#                             Returns true if the specified section exists in the 
#                             current confiuration.
# $id->config_key_exists($$)  Returns true if the specified section and key exists.
# $id->get_cv($$)             Returns the value asociated with the specified section
#                             and key (returns undef if no no matching section is 
#                             found or if no matching key is found in the specified 
#                             section (Arguments: section,key).
# $id->get_cv_test($$$)       Same as above, but for specified test number (defaults to
#                             current test if no is specified).
# $id->get_cv_list($$)        Same as get_cv($$) except it returns a list of elements,
#                             the list is the return value of get_cv($$) split up along any
#                             commas. If the configuration value does not exist or is empty
#                             get_cv_list($$) returns an empty list.
# $id->change_config_value($$$)
#                             Sets the value asociated with the specified section and
#                             key to the specified value (arguments: section,key,newvalue) 
# $id->modtime()              Returns the modification time for the configuration file
#                             fore the test specified (default is curent test).
# $id->debug()                Sets the debug level for the Rig instance as well as all.
#                             device instances attatched to the rig instance.
#                             If called without arguments returns the current debug 
#                             level.
# $id->init()                 Initialises (or reinitialises) the rig instance. This 
#                             member function loads all device instances which according
#                             to the configuration file needs to be attatched. On succes
#                             any call to error() returns an empty string, otherwise the string
#                             returned contains the name of the offending device which
#                             could not be initialised and caused init to fail / die.
#                             If the init() function is not called within an eval{} guard, it 
#                             simply dies with an error indicating what went wrong.
#                             The init() function is automatically called by the constructor
#                             (within an eval{} gurad).
# $id->webdir([opt $testnr])  Returns the current web directory path.
#                             If an argument is specified (and it is a valid test number)
#                             the directory path for that directory is returned instead.
# $id->homedir()              Returns the home directory path for the rig.
# $id->hometest()             Returns the home test directory path for the rig.
#
#
# Device management.
# Any RFC::Rig instance has a number of device instances asociated. Some of these
# devices are read only devices (datalogging) while others may be control devices.
#
# The types avaliable are:
# -- \begin{longtable}{l}
# -- simplechannel \\
# -- relay \\
# -- gas (the plural of this is gasses) \\
# -- gasgroup \\ 
# -- MFC (Mass flow controler) \\
# -- multiplex (plural multiplexers), used for gas multiplexing. \\
# -- water (water bublers) \\
# -- PSU (Power supplies, including electronic loads) \\
# -- templog (Temperature logging only, not control) \\
# -- tempcontrol (Temperature controller) \\
# -- analog (Analog output devices) \\
# -- filter (virtual device) \\
# -- logic (virtual device) \\
# -- PID (Virtual control device for PID control systems) \\
# -- \end{longtable}
#
# and a list of device types can be obtained by calling:
#
# $id->list_types();
#
# Each of these device types has different member functions, but common for all
# of them is that they must be derived from BaseDevice.pm (which defines the base.
# member functions on which RFC::Rig relies on). To get a device instance of a particular 
# type, call
#
# $id->get_device($type,$name);  
#
# The type must match one of the above defined types (otherwise the function merely
# return undef).
# If no device exist of the type and name specified, undefined is returned instead
# of a device instance, note that this function does NOT initialise or load new
# device instances, it merely returns references to already initialised / loaded 
# devices (This is usually handles by the class constructor/init function).
#
# It is also possible to use the init_device function:
#
# $id->init_device($type,$name);
# 
# This function tries to initialise/load the device if it does not already exist.
#
# It is possible to get lists of devices and avaliable devices (that is all possible
# devices of a type, irrespectively of it is loaded or not) as well as supported 
# device types by using the following functions:
#
# $id->list_types()            Returns a list of supported device types.
# $id->list_types_nofilter()   Returns a list of supported device types but excludes 
#                              filter types.
# $id->get_uiname($)           Returns a human readable string describing the
#                              specified device type (intended for use on the UI).
# $id->list_devices($)         Lists the devices of the specified type.
# $id->list_devices_control($) Lists the enabled devices of the specified type which 
#                              is not readonly.
# $id->list_names($)           Lists the devices of the specified type which will be
#                              explicitly initialised during rig initialisation.
#                              Note that this list does not include implicit devices
#                              or devices initialised as part of other devices!
# $id->list_names_tag($)       Returns the tag name used for controling the
#                              enabled status for the devices of the specified
#                              type (an enabled device gets explicitly initialised).
# $id->avaliable_devices($)    Lists the possible devices of the specified type.
# $id->avaliable_devices_nofilter($)
#                              Lists the possible devices of the specified type but 
#                              excludes filter devices which may masquerade.
# $id->device_sectionname($)   Returns the cofiguration section base name (identifier)
#                              for the device type in question. returns 'channel_' if
#                              called with the 'simplechannnel' argument for instance
#                              as the simple channel 'cell_voltage' is configured  in
#                              the 'channel_cell_voltage' section.
# $id->get_devce($$)           Returns a device instance of the specified type and name.
# $id->register_device($$)     Attatches the specified device instance to the rig instance. 
#                              Note duplicate names not allowed. (results in an error).
#                              Argu,ents are: type,name.
# $id->set_cache($$$)          Saves a value for the specified device type and name.
#                              Arguments: Device type, Device name, Value.
# $id->get_cache($$)           Gets the cached value for the specified device type and name.
#                              Returns undef if caching is not allowed or if no value was 
#                              found for the type and name specified.
# $id->store_cache()           Stores the cached values in a file. Is called automatically
#                              by the RFC::Rig destructor.
# $id->load_cache()            Loads the stored cache values.
# $id->clear_cache()           Clears the persistent cache.
# $id->allow_cache([$])        Sets or gets if persistent caching is to be allowed.
#                              Default is to not allow caching. If caching is disallowed
#                              All calls to get_cache() will return undef;
#                              Notice that only if the 'allow_caching' key in the 'IV_control'
#                              is set to 'Yes' is caching allowed to be turned on
#                              (That is, bu setting 'allow_caching' to no disables all 
#                              persistent caching durring iV curves irrespective of device 
#                              configuration).
# $id->list_caching()          Returns a list of devices for which caching is enabled.
#                              Notice that filter devicec can never be caching.
# 
# For all of the above functions the specified type must match one of the types returned
# by $id->list_types() or the functions returns undef.
#
#
# A number of control functions are also defined. They each operate on one 
# of the device instances (depending on type). 
# 
# $id->set_temp($)            Sets the specified temperature setpoint for the specified
#                             device name (if no device name is specified and only one.
#                             temperatue control device is initialised, that device is 
#                             used).
# $id->set_ramp($)            Same as above, but for temperature ramprate (C/hour). 
# $id->current($)             Sets the specified DC curent for the specified PSU
#                             (If no PSU name is specified and only one PSU devce is
#                             initialised, that one is used).
# $id->voltage($)             Same as above, but for the DC voltage.
# $id->get_sessionlog($)      Returns the testlog for the specified test (default is
#                             the current test if no test is specified).
# $id->get_errorlog($)        Returns the errorlog for the specified test (default is
#                             the current test if no test is specified).
# $id->name()                 Returns the Rig name (the integer passed to the RFC::Rig
#                             constructor).
# $id->session()              Returns the current test number.
# $id->proglog($[$])          Appends the specified string to the rig proglog for the
#                             currnet test. If an atdditional argument is specified it
#                             is assumed to be the username of the originating user and
#                             the string is appended with this information, if not a
#                             generic user identification is appended including terminal
#                             of origin (usually 'rigX' on /dev/pts/N). In addition an 
#                             entry is logged in the tracelog in this case.
# $id->errorlog($[$])         Appends the specified string to the rig errorlog for the
#                             currnet test similar to proglog. Note that all logged errors 
#                             are automatically appended to the proglog.
# $id->tracelog($)            Appends the specified string to the rig tracelog for the
#                             currnet test. The logged string includes complete stack 
#                             backtrace
# $id->print_setup()          Prints out the complete current setup including all member 
#                             functions and data fields (uses Class::Inspector).
#                             Overloads the print_setup() function in Debug.pm.
# $id->test_list()            Returns a list of all devices. The returnes list is a 
#                             array of hashes (with each hash containing id and name of
#                             a specific device).
# $id->print_config()         Prints out the complete rig configuration (including all
#                             loaded devices, either explicitly or implicitly).
# $id->check_run()            Forces an error if the current user number does not match 
#                             the rig number (as determined by RFC::Main::get_user_number)
# $id->inc_test($$)           Increases the test number and reinitialises the rig instance.
#                             The aruments are the additional test information and the 
#                             name of the user initialising a net test.
#
# -- \subsection{Private member functions}
#
# Each RFC::Rig instance also has the following private member functions. 
#
# Perl does not enforce public / private declerations however, so the functions are avaliable.
# The private member functions is not intended for normal use, only for advanced use by system 
# maintainers! 
#
# $id->load_session_nr()      Loads the session number from the session number file.
# $id->create_dirs()          Checks the current session and creates the directories
#                             asociated with the test if nescesarry.
#
# Each device type has four handling functins described for the simplechannel class
# below. The four types of handling functions for device type TYPE are:
#
# -- \begin{longtable}{l}
# --  $id->TYPEs(); \\
# --  $id->TYPE(); \\
# --  $id->init_TYPE(); \\
# --  $id->avaliable_TYPEs(); \\
# --  $id->register_TYPE($); \\
# -- \end{longtable}
#
# $id->simplechannels()       Returns a list of the current simplechannels.
# $id->simplechannel($)       Returns the Simplechannel class instance with the 
#                             specified name (if it exists, undef otherwise).
# $id->init_simplechannel($)  Attempts to initialise a new simplechannel device
#                             with the specified name, returns the device instance created.
#                             Forces an error if the nescesarry configuration
#                             section is not found in the configuration file.
# $id->avaliable_simplechannels()
#                             Returns a list of possible simplechannels (including 
#                             disabled simplechannels).
# $id->register_simplechannel($)
#                             Attatches the specified Simplechanel device instance 
#                             to the rig instance. Note duplicate names not allowed.
#                             (results in an error).
#
# Eeach type also has a constructor wraper function asociated with it which
# is a member function of the rig instance:
# 
# -- \begin{longtable}{l}
# -- $id->new_AnalogOut($) \\
# -- $id->new_RELAY($) \\
# -- $id->new_SimpleChannel($) \\
# -- $id->new_PSU($) \\
# -- $id->new_Multiplexer($) \\
# -- $id->new_MFC($) \\
# -- $id->new_TEMP($) \\
# -- $id->new_GAS($) \\
# -- $id->new_Water($) \\
# -- $id->new_Templog($) \\
# -- $id->new_GasGroup($) \\
# -- $id->new_filter($) \\
# -- $id->new_PID($) \\
# -- $id->new_Alert($) \\
# -- \end{longtable}
#
# The new_device functions are wrapper functions for the various Device.pm 
# factory functions. They are used by $rig->init() as well as to obtain an 
# device node for a not enabled device (For testing purposes when installing 
# a ned hwardware/logical device)                      
#
###################################################################

use strict;
use warnings;
use RFC::Header;
use RFC::Main qw();
use RFC::Device;
use ElchemeaConfig;
use SemaforeFile;
use Debug;
use File::Touch;
use POSIX qw();
our @ISA = qw(Debug);
use Fcntl qw(:DEFAULT :flock);
use Class::Inspector;
use File::Copy;

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new();
    my $n = shift;
    if ($n =~ /(\w+)/)
    {
	$self->{'_name'} = $1;
    }
    bless $self,$class;
    $self->{'_session'} = 0;
    $self->{'_oldsession'} = undef;
    $self->{'_config_old'} = [];
    $self->{'_config'} = undef;
    $self->{'_configlock'} = '/home/celltest/RFCRig_'.$self->{'_name'}.'_config.lock';
    $self->{'_userlog'} = undef;
    $self->{'_webtest'} = undef;
    $self->{'_hometest'} = undef;
    $self->{'_riglock'} = '/home/celltest/RFCRig_'.$self->{'_name'}.'.lock';
    $self->{'_tracelock'} = '/home/celltest/RFCRig_'.$self->{'_name'}.'_trace.lock';
    $self->{'_controllock'} = '/home/celltest/RFCRig_'.$self->{'_name'}.'_control.lock';
    $self->{'_info'} = undef;
    $self->{'_trace'} = undef;
    $self->{'_kanaldata'} = [];
    unless (-e $self->{'_riglock'})
    {
	sysopen(FH, $self->{'_riglock'}, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $self->{'_riglock'} $!";
        chmod 0666, $self->{'_riglock'};
        CORE::close FH;
    }
    unless (-e $self->{'_controllock'})
    {
	sysopen(FH, $self->{'_controllock'}, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $self->{'_controllock'} $!";
        chmod 0666, $self->{'_controllock'};
        CORE::close FH;
    }
    my $fn = $self->{'_controllock'};
    sysopen($self->{'_lock_control'}, $fn, O_WRONLY | O_CREAT) or $self->die("Error: Could not open lockfile '$fn' $!");
    $self->{'_tempdatafile'} = SemaforeFile->new('/tmp/rig_'.$self->name.'_data.txt',$self->riglock);
    $self->{'_tempdatafile_stored'} = 1;
    $self->{'_error'} = '';
    $self->{'_kanaldata'} = [];
    $self->{'_TEMPLOG'} = {};
    $self->{'_TEMPCONTROL'} = {};
    $self->{'_GAS'} = {};
    $self->{'_GASGROUP'} = {};
    $self->{'_MFC'} = {};
    $self->{'_WATER'} = {};
    $self->{'_SIMPLE'} = {};
    $self->{'_RELAY'} = {};
    $self->{'_PSU'} = {};
    $self->{'_MULTI'} = {};
    $self->{'_AOUT'} = {};
    $self->{'_FILTER'} = {};
    $self->{'_PID'} = {};
    $self->{'_LOGIC'} = {};
    $self->{'_ALERT'} = {};
    $self->{'_temp_kanaldata'} = {};
    $self->{'_transaction'} = 0;
    $self->{'_types'} = ['simplechannel','relay','templog','tempcontrol','MFC','water','gas','gasgroup','PSU','analog','multiplex','filter','PID','logic','Alert'];
    $self->{'_types_nofilter'} = ['simplechannel','relay','templog','tempcontrol','MFC','water','gas','gasgroup','PSU','analog','multiplex','PID','logic','Alert'];
    $self->{'_uinames'} = {
	'simplechannel' => 'Simple channel',
	'relay' => 'Control relay',
	'templog' => 'Temperature log',
	'tempcontrol' => 'Temperature control',
	'MFC' => 'Mass flow controler',
	'water' => 'Water bubbler',
	'gas' => 'Gas',
	'gasgroup' => 'Gas group',
	'PSU' => 'DC power supply',
	'analog' => 'Analog output',
	'multiplex' => 'Gas multiplexer',
	'filter' => 'Filter device (virtual)',
	'logic' => 'Logic device (virtual)',
	'Alert' => 'Alert device (virtual)',
	'PID' => 'PID control device (virtual)',
    };
    $self->{'_sectionnames'} = {
	'simplechannel' => 'channel_',
	'relay' => 'relay_',
	'templog' => 'temp_',
	'tempcontrol' => 'TControl_',
	'MFC' => 'controler_',
	'water' => 'water_',
	'gas' => 'gas_',
	'gasgroup' => 'gas_group_',
	'PSU' => 'PSU_',
	'analog' => 'analog_out_',
	'multiplex' => 'gas_multiplexer_',
	'filter' => 'filter_',
	'logic' => 'Logic_',
	'Alert' => 'Alert_',
	'PID' => 'PID_',
    };
    unless (-e $self->{'_configlock'})
    {
	sysopen(FH, $self->{'_configlock'}, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $self->{'_configlock'} $!";
        chmod 0666, $self->{'_configlock'};
        CORE::close FH;
    }
    if (@_)
    {
	$self->debug(shift);
    }
    ### Add check for valid rig numbers ###
    $self->die("Rig '".$self->name."' does not exist!") unless RFC::Main::check_rig_nr($self->name);
    $self->{'_iv_in_progress'} = SemaforeFile->new($RFC::Header::publicdir.'rig'.$self->name.'/iv_in_progress',$self->{'_configlock'});
    my $cache_fn = $self->homedir.'/cache';
    unless (-e $cache_fn)
    {
	touch($cache_fn);
	chmod 0666,$cache_fn;
    }
    $self->{'_cache_file'} = SemaforeFile->new($cache_fn,$self->{'_configlock'});
    $self->{'_cache'} = {};
    $self->{'_cache_modified'} = 0;
    $self->{'_allow_cache'} = 0;
    eval {
	$self->init;
    };
    if ($@)
    {
	my $errstr = 'Init failed: '.$@;
	chomp $errstr;
	$errstr .= ', Last attempted device initialisation was: '.$self->{'_error'};
	$self->{'_error'} = $errstr;
	warn $errstr,"\n";
    }
    return $self;
}

return 1;


sub list_caching
{
    my $self = shift;
    my @list = ();
    foreach my $type ($self->list_types_nofilter)
    {
	foreach my $name ($self->avaliable_devices($type))
	{
	    my $dev = $self->get_device($type,$name);
	    push @list,$dev if ($dev && $dev->is_caching());
	}
    }
    return @list;
}

sub allow_cache
{
    my $self = shift;
    if (@_)
    {
	my $test = $self->get_cv('IV_control','allow_caching');
	if ($test && $test =~ /yes/i)
	{
	    $self->{'_allow_cache'} = shift;
	}
    }
    return $self->{'_allow_cache'};
}

sub store_cache()
{
    my $self = shift;
    return undef unless ($self->{'_cache_modified'});
    my @list = ();
    $self->writedebug("Storing persistent cache on disk") if ($self->debug);
    foreach my $key (keys %{$self->{'_cache'}})
    {
	push @list,$key.'='.$self->{'_cache'}{$key} if (defined($self->{'_cache'}{$key}));
    }
    $self->{'_cache_file'}->writeline(@list);
    $self->{'_cache_modified'} = 0;
    return undef;
}

sub clear_cache()
{
    my $self = shift;
    $self->writedebug("Clering persistent cache") if ($self->debug);
    $self->{'_cache_file'}->writeline('');
    $self->{'_cache'} = {};
    $self->{'_cache_modified'} = 0;
    return undef;
}

sub load_cache()
{
    my $self = shift;
    $self->{'_cache'} = {};
    $self->writedebug("Loading persistent cache from disk") if ($self->debug);
    foreach my $line ($self->{'_cache_file'}->readlines())
    {
	next unless ($line);
	my ($key,$val) = split(/\=/,$line);
	$self->{'_cache'}{$key} = $val;
    }
    return undef;
}

sub set_cache($$$)
{
    my $self = shift;
    my $class = shift;
    my $name = shift;
    my $value = shift;
    my $key = $class.':'.$name;
    $self->load_cache() unless (%{$self->{'_cache'}});
    $self->writedebug("Setting cache value '$value' for device '$name' of type '$class'") if ($self->debug);
    $self->{'_cache'}{$key} = $value;
    $self->{'_cache_modified'} = 1;
    return $value;
}

sub get_cache($$)
{
    my $self = shift;
    return undef unless ($self->{'_allow_cache'});
    my $class = shift;
    my $name = shift;
    my $key = $class.':'.$name;
    $self->load_cache() unless (%{$self->{'_cache'}});
    $self->writedebug("retrieving cache value for device '$name' of type '$class'") if ($self->debug);
    return $self->{'_cache'}{$key};
}

sub iv_in_progress()
{
    my $self = shift;
    return $self->{'_iv_in_progress'};
}

sub is_iv()
{
    my $self = shift;
    return $self->{'_iv_in_progress'}->exist;
}

sub iv_start()
{
    my $self = shift;
    my $status = 0;
    $self->{'_iv_in_progress'}->lock_ex;
    unless ($self->{'_iv_in_progress'}->exist)
    {
	$self->{'_iv_in_progress'}->writeline('');
	$status = 1;
    }
    $self->{'_iv_in_progress'}->lock_un;
    return $status;
}

sub iv_stop()
{
    my $self = shift;
    $self->{'_iv_in_progress'}->lock_ex;
    unlink $self->{'_iv_in_progress'}->filename if ($self->{'_iv_in_progress'}->exist);
    $self->{'_iv_in_progress'}->lock_un;
}

sub writedebug
{
    my $self = shift;
    eval {
	$self->tracelog(@_);
    };
    if ($@)
    {
	my $str = $@;
	chomp $str;
	my $mess = join("\n",@_);
	warn "Error: '$str' on atempt to write '$mess' to tracelog!\n";
    }
    return $self->SUPER::writedebug(@_);
}

sub homedir()
{
    my $self = shift;
    return $userdir.'rig'.$self->name.'/';
}

sub hometest()
{
    my $self = shift;
    return $self->homedir.$self->name.'test'.$self->session.'/';
}

sub webdir()
{
    my $self = shift;
    my $test = shift;
    $test = $self->session unless ($test && $test >= 0 && $test < $self->session);
    return $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$test.'/';
}


sub lock_control()
{
    my $self = shift;
    flock($self->{'_lock_control'},LOCK_EX);
}

sub unlock_control()
{
    my $self = shift;
    flock($self->{'_lock_control'},LOCK_UN);
}

sub die($)
{
    my $self = shift;
    my $str = shift;
    if ($self->debug)
    {
	CORE::warn("\n\n Object trace\n\n");
	CORE::warn($self->print_setup."\n");
	foreach my $key ('_MULTI','_MFC','_RELAY','_SIMPLE','_WATER','_GAS','_TEMPLOG','_GASGROUP','_PSU','_AOUT','_TEMPCONTROL','_FILTER','_PID','_LOGIC','_ALERT')
	{
	    CORE::warn('****** '.$key.' ******'."\n");
	    foreach my $k (keys %{$self->{$key}})
	    {
		CORE::warn($self->{$key}{$k}->print_setup."\n");
	    }
	}
    }
    $self->SUPER::die($str);

}

sub config()
{
    my $self = shift;
    return $self->{'_config'};
}

sub begin()
{
    my $self = shift;
    if ($self->{'_transaction'})
    {
	$self->die("A new transaction can not be initiated while inside anoter transaction!");
    }
    $self->{'_config'}->begin();
    $self->{'_transaction'} = 1;
}

sub commit()
{
    my $self = shift;
    unless ($self->{'_transaction'})
    {
	$self->writedebug("Error: Attempting to commit outside a transaction!",1);
    }
    my $res = $self->{'_config'}->commit();
    $self->die($self->{'_config'}->error()) unless ($res);
    $self->{'_transaction'} = 0;
}

sub rollback()
{
    my $self = shift;
    unless ($self->{'_transaction'})
    {
	$self->writedebug("Error: Attempting to rolback outside a transaction!",1);
    }
    $self->{'_config'}->rollback();
    $self->{'_transaction'} = 0;
}

sub error()
{
    my $self = shift;
    return $self->{'_error'};
}

sub load()
{
    my $self = shift;
    ### Load existing data into _temp_kanaldata ###
    my @lines = $self->{'_tempdatafile'}->readlines;
    foreach my $line (@lines)
    {
	my ($name,$time,$data) = split("\t",$line);
	my %h = ('time' => $time, 'data' => $data);
	$self->{'_temp_kanaldata'}{$name} = \%h;	
    }
}

sub store()
{
    my $self = shift;
    my @list = ();
    foreach my $key (keys %{$self->{'_temp_kanaldata'}})
    {
	push @list,join("\t",$key,$self->{'_temp_kanaldata'}{$key}{'time'},$self->{'_temp_kanaldata'}{$key}{'data'});
    }
    my $file = $self->{'_tempdatafile'}->filename;
    unless ($file  && -e $file)
    {
	sysopen(FH, $file, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $file $!";
        chmod 0666, $file;
        CORE::close FH;
    }
    $self->{'_tempdatafile'}->writeline(@list);
    $self->{'_tempdatafile_stored'} = 1;
}

sub get_temp_kanaldata($)
{
    my $self = shift;
    my $key = shift;
    return (undef,undef) unless (exists($self->{'_temp_kanaldata'}{$key}));
    return ($self->{'_temp_kanaldata'}{$key}{'time'},$self->{'_temp_kanaldata'}{$key}{'data'});
}

sub riglock()
{
    my $self = shift;
    return $self->{'_riglock'};
}

sub check_warning($)
{
    my $self = shift;
    my $type = shift;
    my $l = $self->{'_maillock'}->lock_sh_nb;
    my $count = 0;
    while (!$l && $count++ < 3)
    {
	sleep(1);
	$l = $self->{'_maillock'}->lock_sh_nb;
    }
    if ($l)
    {
	my @list = $self->{'_maillock'}->readlines;
	foreach (@list)
	{
	    next unless ($_ eq $type);
	    $self->{'_maillock'}->lock_un;
	    return 1;
	}
    }
    else
    {
	$self->writedebug("Warning! Could not lock '".$self->{'_maillock'}->filename."'",1);
	return undef;
    }
    $self->{'_maillock'}->lock_un;
    return 0;
}

sub set_warning($)
{
    my $self = shift;
    my $type = shift;
    return 0 if ($self->check_warning($type));
    my $l = $self->{'_maillock'}->lock_ex_nb;
    if ($l)
    {
	$self->{'_maillock'}->append($type);
	$self->{'_maillock'}->lock_un;
	return 1;
    }
    else
    {
	$self->writedebug("Warning! Could not exclusively lock '".$self->{'_maillock'}->filename."'",1);
	return 0;
    }
    return 0;
}

sub warning_mail($$)
{
    my $self = shift;
    my $type = shift;
    my $message = shift;
    chomp $message;
    return 0 unless ($self->set_warning($type));
    warn($message,"\n");
    $self->errorlog($message,'warning_system');
    my $str = $self->get_cv('main','warning_mails');
    $str = RFC::Main::get_cv('global','system_mail_users') unless ($str);
    $str = RFC::Main::get_cv('global','errormails') unless ($str);
    if ($str)
    {
	foreach (split(/\s*\,\s*/,$str))
	{
	    if (@_)
	    {
		&RFC::Main::mail($_,$message,@_);
	    }
	    else
	    {
		&RFC::Main::mail($_,$message);
	    }
	}
    }
    return 1;
}

sub unset_warning($)
{
    my $self = shift;
    my $type = shift;
    return 0 unless ($self->check_warning($type));
    my $ok = 0;
    my $l = $self->{'_maillock'}->lock_ex_nb;
    if ($l)
    {
	my @list = $self->{'_maillock'}->readlines;
	my @res = ();
	foreach (@list)
	{
	    unless ($_ eq $type)
	    {
		push @res,$_;
		next;
	    }
	    $ok = 1;
	}
	$self->{'_maillock'}->writeline(@res);
	$self->{'_maillock'}->lock_un;
	return $ok;
    }
    else
    {
	$self->writedebug("Warning! Could not exclusively lock '".$self->{'_maillock'}->filename."'",1);
	return 0;	
    }
}

sub kanaldata
{
    my $self = shift;
    if (@_)
    {
	while (@_)
	{
	    my $str = shift;
	    push @{$self->{'_kanaldata'}},$str;
	    ### Store data in temp kanaldata hash ###
	    my @l = split(/\s+/,$str);
	    if (@l > 2)
	    {
		my %h = ('time'=>$l[0],'data'=>$l[-1]);
		$self->{'_temp_kanaldata'}{$l[1]} = \%h;
		$self->{'_tempdatafile_stored'} = 0;
	    }
	}
	return undef;
    }
    my %h = ();
    #### Remove duplicates ###
    foreach my $k (@{$self->{'_kanaldata'}})
    {
	my @l = split(/\s+/,$k);
	$h{$l[1]} = $k;
    }
    my @out = ();
    foreach my $k (sort {$a cmp $b} keys %h)
    {
	push @out,$h{$k};
    }
    return join("\t",@out);
}

sub readstring()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug) if ($self->debug);
    }
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $logall = 0;
    my $test = $self->get_cv('main','logall');
    $logall = 1 if ($test && $test =~ /yes/i );
    my @arr = ();
    my $dummydev = RFC::BaseDevice->new('dummy',$self);
    ### Notice. do not include filters or logic devices as they are included in the other class type lists already! ###
    ### Also exclude alert devices as the underlying device in thos should already be handles                       ###
    foreach my $key ('_WATER','_TEMPLOG','_GAS','_SIMPLE','_RELAY','_GASGROUP','_AOUT','_TEMPCONTROL','_PSU','_MFC','_MULTI','_PID')
    {
	foreach my $k (sort {$a cmp $b} keys %{$self->{$key}})
	{
	    my $ignore = $self->{$key}{$k}->readstring_ignore;
	    $self->{$key}{$k}->readstring_ignore(0) if ($logall);
	    eval {
		my $str = $self->{$key}{$k}->readstring($data);
		push @arr,$str if defined($str);
	    };
	    if ($@)
	    {
		&RFC::Main::errorlog($@.' on readstring for device "'.$k.'" of type "'.$key.'" on rig '.$self->name);
		#### Create dummy value if device dies ###
		my $name = $self->{$key}{$k}->name;
		my @t = $dummydev->get_time;
		my $t = join(':',@t);
		my $str = join(' ',$name,$t,time,-32768);
		push @arr,$str;
		### Log errorstring in kanaldata
		$str = time . ' '. $name .' '.$@;
		$self->kanaldata($str);
	    }
	    $self->{$key}{$k}->readstring_ignore($ignore) if ($logall);
	}
    }
    my @t = $dummydev->get_time();
    my $t = join(':',@t);
    push @arr,join(' ','VERSION',$t,time,$RFC::Header::VERSION);
    $self->store();
    $self->writedebug("\n".$data->stat()) if ($self->debug > 3);
    return join(' ',@arr);
}

sub list_sections
{
    my $self = shift;
    my @list = $self->{'_config'}->get_sections();
    if (@_)
    {
	my $name = shift;
	my @l = ();
	foreach my $item (@list)
	{
	    if ($item =~ /^$name(\w+)/)
	    {
		push @l,$1;
	    }
	}
	@list = @l;
    }
    return () unless (@list);
    return sort {$a cmp $b} @list;
}


sub avail_filter_type($)
{
    my $self = shift;
    my $type = shift;
    return () unless ($type);
    my @list = ();
    foreach my $name ($self->list_sections('filter'))
    {
	my $c = $self->get_cv('filter'.$name,'class');
	my $t = $self->get_cv('filter'.$name,'device_type');
	if ($c && $c eq 'Typecast')
	{
	    $t = $self->get_cv('filter'.$name,'output_type');
	    $t = 'gas' unless ($t);
	}
	else
	{
	    $t = 'simplechannel' unless ($t);
	}
	my $n = $name;
	$n =~ s/^_//;
	push @list,$n if ($type eq $t);
    }
    return @list;
}

sub Logics()
{
    my $self = shift;
    return keys %{$self->{'_LOGIC'}};
}

sub Logic($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_LOGIC'}{$name} if (exists($self->{'_LOGIC'}{$name}));
    return undef;
}

sub avaliable_Logics()
{
    my $self = shift;
    my @list = $self->avail_filter_type('logic');
    push @list,$self->list_sections('Logic_');
    return @list;
}

sub avaliable_Logics_nofilter()
{
    my $self = shift;
    my @list = ();
    push @list,$self->list_sections('Logic_');
    return @list;
}

sub register_Logic($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->Logics)
    {
	$self->die('Error: Duplicate Logic '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_LOGIC'}{$device->name} = $device;
    return $device;
}

sub Alerts()
{
    my $self = shift;
    return keys %{$self->{'_ALERT'}};
}

sub Alert($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_ALERT'}{$name} if (exists($self->{'_ALERT'}{$name}));
    return undef;
}

sub avaliable_Alerts()
{
    my $self = shift;
    my @list = $self->avail_filter_type('Alert');
    push @list,$self->list_sections('Alert_');
    return @list;
}

sub avaliable_Alerts_nofilter()
{
    my $self = shift;
    my @list = ();
    push @list,$self->list_sections('Alert_');
    return @list;
}

sub register_Alert($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->Alerts)
    {
	$self->die('Error: Duplicate Alert '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_ALERT'}{$device->name} = $device;
    return $device;
}

sub PIDs()
{
    my $self = shift;
    return keys %{$self->{'_PID'}};
}

sub PID($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_PID'}{$name} if (exists($self->{'_PID'}{$name}));
    return undef;
}

sub avaliable_PIDs()
{
    my $self = shift;
    my @list = $self->avail_filter_type('PID');
    push @list,$self->list_sections('PID_');
    return @list;
}

sub avaliable_PIDs_nofilter()
{
    my $self = shift;
    my @list = ();
    push @list,$self->list_sections('PID_');
    return @list;
}

sub register_PID($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->PIDs)
    {
	$self->die('Error: Duplicate PID '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_PID'}{$device->name} = $device;
    return $device;
}

sub simplechannels()
{
    my $self = shift;
    return keys %{$self->{'_SIMPLE'}};
}

sub simplechannel($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_SIMPLE'}{$name} if (exists($self->{'_SIMPLE'}{$name}));
    return undef;
}

sub avaliable_simplechannels()
{
    my $self = shift;
    my @list = $self->avail_filter_type('simplechannel');
    push @list,$self->list_sections('channel_');
    return @list;
}

sub avaliable_simplechannels_nofilter()
{
    my $self = shift;
    my @list = ();
    push @list,$self->list_sections('channel_');
    return @list;
}

sub register_simplechannel($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->simplechannels)
    {
	$self->die('Error: Duplicate simplechannel '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_SIMPLE'}{$device->name} = $device;
    return $device;
}

sub filters()
{
    my $self = shift;
    return keys %{$self->{'_FILTER'}};
}

sub filter()
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_FILTER'}{$name} if (exists($self->{'_FILTER'}{$name}));
    return undef;
}

sub avaliable_filters()
{
    my $self = shift;
    return $self->list_sections('filter_');
}

sub register_filter($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->filters)
    {
	$self->die('Error: Duplicate filter '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_FILTER'}{$device->name} = $device;
    return $device;
}

sub relays()
{
    my $self = shift;
    return keys %{$self->{'_RELAY'}};
}

sub relay($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_RELAY'}{$name} if (exists($self->{'_RELAY'}{$name}));
    return undef;
}

sub avaliable_relays()
{
    my $self = shift;
    my @list = $self->avail_filter_type('relay');
    push @list, $self->list_sections('relay_');
    return @list;
}

sub avaliable_relays_nofilter()
{
    my $self = shift;
    return $self->list_sections('relay_');
}

sub register_relay($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->relays)
    {
	$self->die('Error: Duplicate relay '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_RELAY'}{$device->name} = $device;
    return $device;
}

sub gasses()
{
    my $self = shift;
    return keys %{$self->{'_GAS'}};
}

sub gas($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_GAS'}{$name} if (exists($self->{'_GAS'}{$name}));
    ### Also include gas groups in gasses ###
    return $self->{'_GASGROUP'}{$name} if (exists($self->{'_GASGROUP'}{$name}));    
    return undef;
}

sub avaliable_gasses()
{
    my $self = shift;
    my @list = $self->list_sections('gas_');
    my @l = $self->avail_filter_type('gas');
    foreach (@list)
    {
	next if (/group_/);
	next if (/multiplexer_/);
	push @l,$_;
    }
    push @l,$self->avaliable_gasgroups();
    return @l;
}

sub avaliable_gasses_nofilter()
{
    my $self = shift;
    my @list = $self->list_sections('gas_');
    my @l = ();
    foreach (@list)
    {
	next if (/group_/);
	next if (/multiplexer_/);
	push @l,$_;
    }
    return @l;
}

sub register_gas($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->gasses)
    {
	$self->die('Error: Duplicate gas '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_GAS'}{$device->name} = $device;
    return $device;
}

sub gasgroups()
{
    my $self = shift;
    return keys %{$self->{'_GASGROUP'}};
}

sub gasgroup($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_GASGROUP'}{$name} if (exists($self->{'_GASGROUP'}{$name}));
    return undef;
}

sub avaliable_gasgroups()
{
    my $self = shift;
    my @list = $self->avail_filter_type('gas_group');
    push @list, $self->list_sections('gas_group_');
    return @list;
}

sub avaliable_gasgroups_nofilter()
{
    my $self = shift;
    return $self->list_sections('gas_group_');
}

sub register_gasgroup($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->gasgroups)
    {
	$self->die('Error: Duplicate gas group '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_GASGROUP'}{$device->name} = $device;
    return $device;
}

sub MFCs()
{
    my $self = shift;
    return keys %{$self->{'_MFC'}};
}

sub MFC($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_MFC'}{$name} if (exists($self->{'_MFC'}{$name}));
    return undef;
}

sub avaliable_MFCs()
{
    my $self = shift;
    my @list = $self->avail_filter_type('MFC');
    push @list, $self->list_sections('controler_');
    return @list;
}

sub avaliable_MFCs_nofilter()
{
    my $self = shift;
    return $self->list_sections('controler_');
}

sub register_MFC($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->MFCs)
    {
	$self->die('Error: Duplicate MFC '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_MFC'}{$device->name} = $device;
    return $device
}

sub multiplexers()
{
    my $self = shift;
    return keys %{$self->{'_MULTI'}};
}

sub multiplex($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_MULTI'}{$name} if (exists($self->{'_MULTI'}{$name}));
    return undef;
}

sub avaliable_multiplexers()
{
    my $self = shift;
    my @list = $self->avail_filter_type('multiplex');
    push @list, $self->list_sections('gas_multiplexer_');
    return @list;
}

sub avaliable_multiplexers_nofilter()
{
    my $self = shift;
    return $self->list_sections('gas_multiplexer_');
}

sub register_multiplex($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->multiplexers)
    {
	$self->die('Error: Duplicate gas multiplexer '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_MULTI'}{$device->name} = $device;
    return $device
}

sub waters()
{
    my $self = shift;
    return keys %{$self->{'_WATER'}};
}

sub water($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_WATER'}{$name} if (exists($self->{'_WATER'}{$name}));
    return undef;
}

sub avaliable_waters()
{
    my $self = shift;
    my @list = $self->avail_filter_type('water');
    push @list, $self->list_sections('water_');
    return @list;
}

sub avaliable_waters_nofilter()
{
    my $self = shift;
    return $self->list_sections('water_');
}

sub register_water($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->waters)
    {
	$self->die('Error: Duplicate water bubler '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_WATER'}{$device->name} = $device;
    return $device;
}

sub PSUs()
{
    my $self = shift;
    return keys %{$self->{'_PSU'}};
}

sub PSU($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_PSU'}{$name} if (exists($self->{'_PSU'}{$name}));
    return undef;
}

sub avaliable_PSUs()
{
    my $self = shift;
    my @list = $self->avail_filter_type('PSU');
    push @list, $self->list_sections('PSU_');
    return @list;
}

sub avaliable_PSUs_nofilter()
{
    my $self = shift;
    return $self->list_sections('PSU_');
}

sub register_PSU($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->PSUs)
    {
	$self->die('Error: Duplicate PSU '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_PSU'}{$device->name} = $device;
    return $device;
}

sub templog($)
{
    my $self = shift;
    return $self->temp(@_);
}

sub templogs()
{
    my $self = shift;
    return $self->temps();
}

sub avaliable_templogs()
{
    my $self = shift;
    return $self->avaliable_temps();
}

sub avaliable_templogs_nofilter()
{
    my $self = shift;
    return $self->avaliable_temps_nofilter();
}

sub register_templog($)
{
    my $self = shift;
    return $self->register_temp(@_);
}

sub temps()
{
    my $self = shift;
    return keys %{$self->{'_TEMPLOG'}};
}


sub temp($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_TEMPLOG'}{$name} if (exists($self->{'_TEMPLOG'}{$name}));
    return undef;
}

sub avaliable_temps()
{
    my $self = shift;
    my @list = $self->avail_filter_type('templog');
    push @list, $self->list_sections('temp_');
    return @list;
}

sub avaliable_temps_nofilter()
{
    my $self = shift;
    return $self->list_sections('temp_');
}

sub register_temp($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->temps)
    {
	$self->die('Error: Duplicate temperature device '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_TEMPLOG'}{$device->name} = $device;
    return $device;
}

sub tempcontrols()
{
    my $self = shift;
    return keys %{$self->{'_TEMPCONTROL'}};
}

sub tempcontrol($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_TEMPCONTROL'}{$name} if (exists($self->{'_TEMPCONTROL'}{$name}));
    return undef;
}

sub avaliable_tempcontrols()
{
    my $self = shift;
    my @list = $self->avail_filter_type('tempcontrol');
    push @list, $self->list_sections('TControl_');
    return @list;
}

sub avaliable_tempcontrols_nofilter()
{
    my $self = shift;
    return $self->list_sections('TControl_');
}

sub register_tempcontrol($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->tempcontrols)
    {
	$self->die('Error: Duplicate temperature controler '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_TEMPCONTROL'}{$device->name} = $device;
    return $device;
}

sub analog($)
{
    my $self = shift;
    my $name = shift;
    return undef unless defined($name);
    return $self->{'_AOUT'}{$name} if (exists($self->{'_AOUT'}{$name}));
    return undef;
}

sub analogs()
{
    my $self = shift;
    return keys %{$self->{'_AOUT'}};
}

sub avaliable_analogs()
{
    my $self = shift;
    my @list = $self->avail_filter_type('analog');
    push @list, $self->list_sections('analog_out_');
    return @list;
}

sub avaliable_analogs_nofilter()
{
    my $self = shift;
    return $self->list_sections('analog_out_');
}

sub register_analog($)
{
    my $self = shift;
    my $device = shift;
    foreach ($self->analogs)
    {
	$self->die('Error: Duplicate analog output '.$device->name." attempted!\n") if ($device->name eq $_);
    }
    $self->{'_AOUT'}{$device->name} = $device;
    return $device;
}



sub modtime($)
{
    my $self = shift;
    my $session = $self->session;
    if (@_)
    {
	$session = shift if (@_);
	return $self->{'_config_old'}[$session]->modtime;	
    }
    else
    {
	return $self->{'_config'}->modtime();
    }
}

sub debug()
{
    my $self = shift;
    $self->SUPER::debug(@_);
    if (@_)
    {
        #### Pass on debug level to all objects honoring this ###
        foreach my $i (@{$self->{'_config_old'}})
        {
            $i->debug($self->{'_debug'}) if ($i && ref($i));
        }
        $self->{'_config'}->debug($self->{'_debug'}) if $self->{'_config'};
        $self->{'_info'}->debug($self->{'_debug'}) if $self->{'_info'};
        $self->{'_trace'}->debug($self->{'_debug'}) if $self->{'_trace'};
        $self->{'_maillock'}->debug($self->{'_debug'}) if $self->{'_maillock'};
        $self->{'_userlog'}->debug($self->{'_debug'}) if $self->{'_userlog'};
	$self->{'_tempdatafile'}->debug($self->{'_debug'}) if $self->{'_tempdatafile'};
	### Ignore tempcontrol devices as the Eurotherm class does not support the debug function ###
	foreach my $key ('_MFC','_RELAY','_SIMPLE','_WATER','_GAS','_TEMPLOG','_PSU','_MULTI','_GASGROUP','_TEMPCONTROL','_FILTER','_PID','_LOGIC','_ALERT')
	{
	    foreach my $k (keys %{$self->{$key}})
	    {
		$self->{$key}{$k}->debug($self->{'_debug'});
	    }
	}
    }
    return $self->{'_debug'};
}

sub info
{
    my $self = shift;
    if (@_)
    {
	my $l = $self->{'_info'}->lock_ex_nb();
	unless ($l)
	{
	    sleep(1);
	    $l = $self->{'_info'}->lock_ex_nb();
	}
	if ($l)
	{
	    $self->{'_info'}->append(@_);
	    $self->{'_info'}->lock_un();
	}
	else
	{
	    warn "Could not lock info file for writing! Message '".join("\n",@_)."' Could not be appended!\n";
	}
    }
    my $res = '';
    my $l = $self->{'_info'}->lock_sh_nb();
    unless ($l)
    {
	sleep(1);
	$l = $self->{'_info'}->lock_sh_nb();
    }
    if ($l)
    {
	$res = join("\n",$self->{'_info'}->readlines);
	$self->{'_info'}->lock_un();
    }
    return $res;
}

sub list_types()
{
    my $self = shift;
    return @{$self->{'_types'}};
}

sub list_types_nofilter()
{
    my $self = shift;
    return @{$self->{'_types_nofilter'}};
}

sub device_sectionname($)
{
    my $self = shift;
    my $type = shift;
    return undef unless ($type);
    return $self->{'_sectionnames'}{$type} if (exists($self->{'_sectionnames'}{$type}));
    return undef;
}

sub get_list_config($$)
{
    ### Deprecated, do not use ###
    my $self = shift;
    my $section = shift;
    my $key = shift;
    return $self->get_cv_list($section,$key);
}

sub get_uiname($)
{
    my $self = shift;
    my $type = shift;
    return '' unless ($type);
    return $self->{'_uinames'}{$type} if (exists($self->{'_uinames'}{$type}));
    return '';
}

sub list_names($)
{
    my $self = shift;
    my $type = shift;
    return undef unless ($type);
    return $self->get_cv_list('main','simple_channel_names') if ($type eq 'simplechannel');
    return $self->get_cv_list('main','relay_names') if ($type eq 'relay');
    return $self->get_cv_list('main','temp_names') if ($type eq 'templog');
    return $self->get_cv_list('main','temperature_control_names') if ($type eq 'tempcontrol');
    return $self->get_cv_list('main','automatic_flowcontrolers') if ($type eq 'MFC');
    return $self->get_cv_list('main','water_names') if ($type eq 'water');
    return $self->get_cv_list('main','gas_names') if ($type eq 'gas');
    return $self->get_cv_list('main','gas_group_names') if ($type eq 'gasgroup');
    return $self->get_cv_list('main','PSU_names') if ($type eq 'PSU');
    return $self->get_cv_list('main','analog_controlers') if ($type eq 'analog');
    return $self->get_cv_list('main','gas_multiplexers') if ($type eq 'multiplex');
    return $self->get_cv_list('main','filter_names') if ($type eq 'filter');
    return $self->get_cv_list('main','logic_names') if ($type eq 'logic');
    return $self->get_cv_list('main','PID_names') if ($type eq 'PID');
    return $self->get_cv_list('main','Alert_names') if ($type eq 'Alert');
    return ();
}

sub list_names_tag($)
{
    my $self = shift;
    my $type = shift;
    return undef unless ($type);
    return 'simple_channel_names' if ($type eq 'simplechannel');
    return 'relay_names' if ($type eq 'relay');
    return 'temp_names' if ($type eq 'templog');
    return 'temperature_control_names' if ($type eq 'tempcontrol');
    return 'automatic_flowcontrolers' if ($type eq 'MFC');
    return 'water_names' if ($type eq 'water');
    return 'gas_names' if ($type eq 'gas');
    return 'gas_group_names' if ($type eq 'gasgroup');
    return 'PSU_names' if ($type eq 'PSU');
    return 'analog_controlers' if ($type eq 'analog');
    return 'gas_multiplexers' if ($type eq 'multiplex');
    return 'filter_names' if ($type eq 'filter');
    return 'logic_names' if ($type eq 'logic');
    return 'PID_names' if ($type eq 'PID');
    return 'Alert_names' if ($type eq 'Alert');
    return undef;
}

sub get_device($$)
{
    my $self = shift;
    my $type = shift;
    my $name = shift;
    return undef unless ($type && $name);
    return $self->simplechannel($name) if ($type eq 'simplechannel');
    return $self->relay($name) if ($type eq 'relay');
    return $self->temp($name) if ($type eq 'templog');
    return $self->tempcontrol($name) if ($type eq 'tempcontrol');
    return $self->MFC($name) if ($type eq 'MFC');
    return $self->water($name) if ($type eq 'water');
    return $self->gas($name) if ($type eq 'gas');
    return $self->gasgroup($name) if ($type eq 'gasgroup');
    return $self->PSU($name) if ($type eq 'PSU');
    return $self->analog($name) if ($type eq 'analog');
    return $self->multiplex($name) if ($type eq 'multiplex');
    return $self->filter($name) if ($type eq 'filter');
    return $self->Logic($name) if ($type eq 'logic');
    return $self->PID($name) if ($type eq 'PID');
    return $self->Alert($name) if ($type eq 'Alert');
    return undef;
}

sub register_device($$)
{
    my $self = shift;
    my $type = shift;
    my $name = shift;
    return undef unless ($type && $name);
    return $self->register_simplechannel($name) if ($type eq 'simplechannel');
    return $self->register_relay($name) if ($type eq 'relay');
    return $self->register_temp($name) if ($type eq 'templog');
    return $self->register_tempcontrol($name) if ($type eq 'tempcontrol');
    return $self->register_MFC($name) if ($type eq 'MFC');
    return $self->register_water($name) if ($type eq 'water');
    return $self->register_gas($name) if ($type eq 'gas');
    return $self->register_gasgroup($name) if ($type eq 'gasgroup');
    return $self->register_PSU($name) if ($type eq 'PSU');
    return $self->register_analog($name) if ($type eq 'analog');
    return $self->register_multiplex($name) if ($type eq 'multiplex');
    return $self->register_filter($name) if ($type eq 'filter');
    return $self->register_Logic($name) if ($type eq 'logic');
    return $self->register_PID($name) if ($type eq 'PID');
    return $self->register_Alert($name) if ($type eq 'Alert');
    return undef;
}

sub get_device_control($$)
{
    my $self = shift;
    my $type = shift;
    my $name = shift;
    return undef unless ($type && $name);
    my $dev = $self->get_device($type,$name);
    return $dev if ($dev && !$dev->readonly && !$dev->readstring_ignore);
    return undef;
}

sub list_devices($)
{
    my $self = shift;
    my $type = shift;
    return undef unless ($type);
    return $self->simplechannels() if ($type eq 'simplechannel');
    return $self->relays() if ($type eq 'relay');
    return $self->temps() if ($type eq 'templog');
    return $self->tempcontrols() if ($type eq 'tempcontrol');
    return $self->MFCs() if ($type eq 'MFC');
    return $self->waters() if ($type eq 'water');
    return $self->gasses() if ($type eq 'gas');
    return $self->gasgroups() if ($type eq 'gasgroup');
    return $self->PSUs() if ($type eq 'PSU');
    return $self->analogs() if ($type eq 'analog');
    return $self->multiplexers() if ($type eq 'multiplex');
    return $self->filters() if ($type eq 'filter');
    return $self->Logics() if ($type eq 'logic');
    return $self->PIDs() if ($type eq 'PID');
    return $self->Alerts() if ($type eq 'Alert');
    return undef;
}

sub list_devices_control($)
{
    my $self = shift;
    my $type = shift;
    return undef unless ($type);
    my @list = $self->list_devices($type);
    my @res = ();
    foreach my $n (@list)
    {
	my $dev = $self->get_device($type,$n);
	if ($dev && !$dev->readonly && !$dev->readstring_ignore)
	{
	    push @res,$n;
	    next;
	}
	push @res,$n if ($type eq 'tempcontrol' && $n eq 'temperature');
	push @res,$n if ($type eq 'PSU' && $n eq 'current');
	push @res,$n if ($type eq 'multiplex');
    }
    return @res;
}

sub avaliable_devices($)
{
    my $self = shift;
    my $type = shift;
    return undef unless ($type);
    return $self->avaliable_simplechannels() if ($type eq 'simplechannel');
    return $self->avaliable_relays() if ($type eq 'relay');
    return $self->avaliable_temps() if ($type eq 'templog');
    if ($type eq 'tempcontrol')
    {
	my @list = $self->avaliable_tempcontrols();
	return @list if (@list);
	return 'temperature' if ($self->config->section_exists('temperature'));
    }
    return $self->avaliable_MFCs() if ($type eq 'MFC');
    return $self->avaliable_waters() if ($type eq 'water');
    return $self->avaliable_gasses() if ($type eq 'gas');
    return $self->avaliable_gasgroups() if ($type eq 'gasgroup');
    if ($type eq 'PSU')
    {
	my @list = $self->avaliable_PSUs();
	return @list if (@list);
	return 'current' if ($self->config->section_exists('current'));
    }
    return $self->avaliable_analogs() if ($type eq 'analog');
    return $self->avaliable_multiplexers() if ($type eq 'multiplex');
    return $self->avaliable_filters() if ($type eq 'filter');
    return $self->avaliable_Logics() if ($type eq 'logic');
    return $self->avaliable_PIDs() if ($type eq 'PID');
    return $self->avaliable_Alerts() if ($type eq 'Alert');
    return undef;
}


sub avaliable_devices_nofilter($)
{
    my $self = shift;
    my $type = shift;
    return undef unless ($type);
    return $self->avaliable_simplechannels_nofilter() if ($type eq 'simplechannel');
    return $self->avaliable_relays_nofilter() if ($type eq 'relay');
    return $self->avaliable_temps_nofilter() if ($type eq 'templog');
    if ($type eq 'tempcontrol')
    {
	my @list = $self->avaliable_tempcontrols_nofilter();
	return @list if (@list);
	return 'temperature' if ($self->config->section_exists('temperature'));
    }
    return $self->avaliable_MFCs_nofilter() if ($type eq 'MFC');
    return $self->avaliable_waters_nofilter() if ($type eq 'water');
    return $self->avaliable_gasses_nofilter() if ($type eq 'gas');
    return $self->avaliable_gasgroups_nofilter() if ($type eq 'gasgroup');
    if ($type eq 'PSU')
    {
	my @list = $self->avaliable_PSUs_nofilter();
	return @list if (@list);
	return 'current' if ($self->config->section_exists('current'));
    }
    return $self->avaliable_analogs_nofilter() if ($type eq 'analog');
    return $self->avaliable_multiplexers_nofilter() if ($type eq 'multiplex');
    return $self->avaliable_filters() if ($type eq 'filter');
    return $self->avaliable_Logics() if ($type eq 'logic');
    return $self->avaliable_PIDs() if ($type eq 'PID');
    return $self->avaliable_Alerts() if ($type eq 'Alert');
    return undef;
}

sub DESTROY
{
    my $self = shift;
    $self->store() unless ($self->{'_tempdatafile_stored'});
    $self->store_cache();
    if ($self->{'_backup'})
    {
	$self->writedebug("Implicitly issuing rollback for incomplete transaction",1);
	$self->rollback;
    }
    ## Below part is needed to force all fiel handles to be closed.....                   ###
    ## needed to make sure that data are stored in devices! (Perl WTF garbage collector!) ### 
    foreach my $key ('_MULTI','_MFC','_RELAY','_SIMPLE','_WATER','_GAS','_TEMPLOG','_GASGROUP','_PSU','_AOUT','_TEMPCONTROL','_FILTER','_PID','_LOGIC','_ALERT')
    {
	### Force destructors in child devices to awoid the Perl garbage collector to mess things up.... ###
	foreach my $k (keys %{$self->{$key}})
	{
	    $self->{$key}{$k}->DESTROY() if ($self->{$key}{$k});
	}
    }
    for my $x (0..$self->session)
    {
	$self->{'_config_old'}[$x]->DESTROY if ($self->{'_config_old'}[$x] && ref($self->{'_config_old'}[$x]));
    }
    foreach my $item ('_maillock','_info','_trace','_userlog','_errlog','_tempdatafile','_cache_file','_iv_in_progress','_config')
    {
	$self->{$item}->DESTROY if ($self->{$item});
    }
    CORE::close($self->{'_lock_control'});
    return;
}


sub init()
{
    my $self = shift;
    $self->{'_kanaldata'} = [];
    $self->{'_TEMPLOG'} = {};
    $self->{'_TEMPCONTROL'} = {};
    $self->{'_GAS'} = {};
    $self->{'_GASGROUP'} = {};
    $self->{'_MFC'} = {};
    $self->{'_WATER'} = {};
    $self->{'_SIMPLE'} = {};
    $self->{'_RELAY'} = {};
    $self->{'_PSU'} = {};
    $self->{'_MULTI'} = {};
    $self->{'_AOUT'} = {};
    $self->{'_FILTER'} = {};
    $self->{'_PID'} = {};
    $self->{'_LOGIC'} = {};
    $self->{'_ALERT'} = {};
    $self->{'_temp_kanaldata'} = {};
    $self->{'_error'} = '';
    $self->load_session_nr;
    $self->create_dirs;
    my $file = '/home/celltest/Gas_rig'.$self->name.'.lock';
    unless (-e $file)
    {
	sysopen(FH, $file, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $file $!";
        chmod 0666, $file;
	CORE::close FH;
    }
    unless (defined($self->{'_oldsession'}) && $self->session == $self->{'_oldsession'})
    {
	$self->die("Initialisation attempted for new session inside a transaction!") if ($self->{'_transaction'});
	$self->{'_config'} = ElchemeaConfig->new($publicdir.'rig'.$self->name.'/configfile',$self->{'_configlock'}); 
	$self->{'_config'}->debug($self->{'_debug'});
	$self->{'_config_old'} = undef;
	for my $x (0..$self->session)
	{
	    my $fn = $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$x.'/configfile';
	    if (-e $fn)
	    {
		### Only store filename and delay ElchemeaConfig initialisation to the config is actually needed ###
		### As in most cases old config values are normally only needed for reporting purposes and not   ###
		### for normal data aquisition or process control                                                ###
		### This sppeds rig initialisation up ad awoid having a number of unnescesarry open file handles ###
		### (As each ElchemeaConfig instance opens the lock file for locking tests as part of the        ###
		### constructor                                                                                  ###
		$self->{'_config_old'}[$x] = $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$x.'/configfile';
	    }
	    else
	    {
		## Use current configuration if no old one found ###
		$self->{'_config_old'}[$x] = $self->{'_config'};
	    }
	}
    }
    $file = $self->homedir.'/maillock.txt';
    unless (-e $file)
    {
	sysopen(FH, $file, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $file $!";
        chmod 0666, $file;
	CORE::close FH;
    }    
    $self->{'_maillock'} = SemaforeFile->new($file,$self->{'_riglock'});
    $self->{'_hometest'} = $userdir.'rig'.$self->name.'/'.$self->name.'test'.$self->session.'/';
    $self->{'_maillock'}->debug($self->{'_debug'});
    $file = $self->{'_webtest'}.'info.txt';
    unless (-e $file)
    {
	sysopen(FH, $file, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $file $!";
        chmod 0666, $file;
	CORE::close FH;
    }
    $self->{'_info'} = SemaforeFile->new($file,$self->{'_riglock'});
    $self->{'_info'}->debug($self->{'_debug'});
    $self->{'_trace'} = SemaforeFile->new($self->{'_webtest'}.'tracelog.txt',$self->{'_tracelock'});
    $self->{'_trace'}->debug($self->{'_debug'});
    $self->writedebug("Initialising user '$self->{'_name'}' configuration") if ($self->debug);
    $self->{'_userlog'} = SemaforeFile->new($self->{'_webtest'}.'proglog',$self->{'_riglock'});
    $self->{'_userlog'}->debug($self->{'_debug'});
    $self->{'_errlog'} = SemaforeFile->new($self->{'_webtest'}.'errorlog',$self->{'_tracelock'});
    $self->{'_errlog'}->debug($self->{'_debug'});
    $self->{'_modtime'} = $self->modtime();
    ### Initialise all filter devices      ####
    my $str = $self->get_cv('main','filter_names');
    $str = '' unless ($str);
    foreach my $name ($self->get_cv_list('main','filter_names'))
    {
	$self->{'_error'} = 'filter device '.$name;
	$self->init_filter($name);
    }
    ### Initialise all filter logic      ####
    $str = undef;
    $str = $self->get_cv('main','logic_names');
    $str = '' unless ($str);
    foreach my $name ($self->get_cv_list('main','logic_names'))
    {
	$self->{'_error'} = 'logic device '.$name;
	$self->init_logic($name);
    }
    ### Start by initialise simple devices ####
    foreach my $name ($self->get_cv_list('main','analog_controlers'))
    {
	$self->{'_error'} = 'Analog output '.$name;
	if ($self->{'_AOUT'}{$name})
	{
	    $self->{'_AOUT'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_analog($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','relay_names'))
    {
	$self->{'_error'} = 'Relay '.$name;
	if ($self->{'_RELAY'}{$name})
	{
	    $self->{'_RELAY'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_relay($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','simple_channel_names'))
    {
	$self->{'_error'} = 'Simple channel '.$name;
	if ($self->{'_SIMPLE'}{$name})
	{
	    $self->{'_SIMPLE'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_simplechannel($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','PSU_names'))
    {
	$self->{'_error'} = 'PSU '.$name;
	### Skip initialisation for already initialised devices ###
	if ($self->{'_PSU'}{$name})
	{
	    $self->{'_PSU'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_PSU($name);
	}
    }
    #### Assign a single default Delta SM-15-100 if no PSU lables found for backwards compatibility ###
    ## Only included for backwards compatibility, and in that case readstring_ignore has to be set  ###
    unless ($self->get_cv_list('main','PSU_names'))
    {
	$self->{'_error'} = 'PSU current ';
	if ($self->{'_config'}->section_exists('current'))
	{
	    unless (exists($self->{'_PSU'}{'current'}))
	    {
		my $dev = $self->new_PSU('current','DefaultDelta'); 
		$self->register_PSU($dev);
		$dev->readstring_ignore(1) if ($dev);
	    }
	}
    }
    ### Then initialise complex devices which may include simple devices implicitly or explictly ###
    foreach my $name ($self->get_cv_list('main','gas_multiplexers'))
    {
	$self->{'_error'} = 'Multiplex '.$name;
	if ($self->{'_MULTI'}{$name})
	{
	    $self->{'_MULTI'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_multiplex($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','automatic_flowcontrolers'))
    {
	$self->{'_error'} = 'MFC '.$name;
	if ($self->{'_MFC'}{$name})
	{
	    $self->{'_MFC'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_MFC($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','temperature_control_names'))
    {
	$self->{'_error'} = 'Temperature controler '.$name;
	if ($self->{'_TEMPCONTROL'}{$name})
	{
	    $self->{'_TEMPCONTROL'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_tempcontrol($name);
	}
    }
    ## Assign a single temperature controler using section 'temperature' if no labels found ###
    ## Only included for backwards compatibility, and in that case readstring_ignore has to ###
    ## be set                                                                               ###
    unless ($self->get_cv_list('main','temperature_control_names'))
    {
	$self->{'_error'} = 'Temperature controler temperature ';
	if ($self->{'_config'}->section_exists('temperature'))
	{
	    my $dev = RFC::Device::new_TEMP($self,'temperature');
            $self->register_tempcontrol($dev);
	    $dev->readstring_ignore(1);
	}
    }
    foreach my $name ($self->get_cv_list('main','gas_names'))
    {
	$self->{'_error'} = 'Gas '.$name;
	### Skip initialisation for already initialised devices ###
	if ($self->{'_GAS'}{$name})
	{
	    $self->{'_GAS'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_gas($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','water_names'))
    {
	$self->{'_error'} = 'Water '.$name;
	if ($self->{'_WATER'}{$name})
	{
	    $self->{'_WATER'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_water($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','temp_names'))
    {
	$self->{'_error'} = 'Temperature log '.$name;
	if ($self->{'_TEMPLOG'}{$name})
	{
	    $self->{'_REMPLOG'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_temp($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','gas_group_names'))
    {
	$self->{'_error'} = 'Gas group '.$name;
	if ($self->{'_GASGROUP'}{$name})
	{
	    $self->{'_GASGROUP'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_gasgroup($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','PID_names'))
    {
	$self->{'_error'} = 'PID '.$name;
	if ($self->{'_PID'}{$name})
	{
	    $self->{'_PID'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_PID($name);
	}
    }
    foreach my $name ($self->get_cv_list('main','Alert_names'))
    {
	$self->{'_error'} = 'Alert '.$name;
	if ($self->{'_ALERT'}{$name})
	{
	    $self->{'_ALERT'}{$name}->readstring_ignore(0);
	}
	else
	{
	    $self->init_Alert($name);
	}
    }
    ### Run init() on all filter devices      ####
    foreach my $name (keys %{$self->{'_FILTER'}})
    {
	$self->{'_error'} = 'Filter device init: '.$name;
	$self->{'_FILTER'}{$name}->init();
    }
    ### Run init() on all logic devices      ####
    foreach my $name (keys %{$self->{'_LOGIC'}})
    {
	$self->{'_error'} = 'Logic device init: '.$name;
	$self->{'_LOGIC'}{$name}->init();
    }
    $self->{'_modtime'} = $self->modtime();
    $self->load();
    ### Reset errorstring as last operation ###
    $self->{'_error'} = '';
    $self->{'_oldsession'} = $self->session;
    return $self;
}

sub init_Alert($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_Alert!\n" unless (defined($name));
    CORE::die "No information found for Alert '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('Alert_'.$name));
    $self->register_Alert($self->new_Alert($name));
    return $self->Alert($name)
}

sub init_PID($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_PID!\n" unless (defined($name));
    CORE::die "No information found for PID '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('PID_'.$name));
    $self->register_PID($self->new_PID($name));
    return $self->PID($name)
}

sub init_analog($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_analog!\n" unless (defined($name));
    CORE::die "No information found for analog output '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('analog_out_'.$name));
    $self->register_analog($self->new_AnalogOut($name));
    return $self->analog($name)
}

sub init_filter()
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_filter!\n" unless (defined($name));
    CORE::die "No information found for filter device '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('filter_'.$name));
    $self->register_filter($self->new_filter($name));
    return $self->filter($name);
}

sub init_logic()
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_logic!\n" unless (defined($name));
    CORE::die "No information found for logic device '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('Logic_'.$name));
    $self->register_Logic($self->new_Logic($name));
    return $self->Logic($name);
}

sub init_relay($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_relay!\n" unless (defined($name));
    CORE::die "No information found for relay '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('relay_'.$name));
    $self->register_relay($self->new_RELAY($name));
    return $self->relay($name);
}

sub init_simplechannel($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_simplechannel!\n" unless (defined($name));
    CORE::die "No information found for simple channel '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('channel_'.$name));
    $self->register_simplechannel($self->new_SimpleChannel($name));
    return $self->simplechannel($name);
}

sub init_PSU($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_PSU!\n" unless (defined($name));
    CORE::die "No information found for PSU '$name'".' for rig '.$self->name,"\n" unless ($self->{'_config'}->section_exists('PSU_'.$name));
    $self->register_PSU($self->new_PSU($name));
    return $self->PSU($name);
}

sub init_multiplex($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_multiplex!\n" unless (defined($name));
    CORE::die "No information found for gas multiplexer '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('gas_multiplexer_'.$name));
    $self->register_multiplex($self->new_Multiplexer($name));
    return $self->multiplex($name);
}

sub init_MFC($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_MFC!\n" unless (defined($name));
    CORE::die "No information found for MFC '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('controler_'.$name));
    $self->register_MFC($self->new_MFC($name));
    return $self->MFC($name);
}

sub init_tempcontrol($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_tempcontrol!\n" unless (defined($name));
    CORE::die "No information found for temperature controler '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('TControl_'.$name));
    $self->register_tempcontrol($self->new_TEMP($name));
    return $self->tempcontrol($name);
}


sub init_gas($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_gas!\n" unless (defined($name));
    CORE::die "No information found for gas '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('gas_'.$name));
    $self->register_gas($self->new_GAS($name));
    return $self->gas($name);
}

sub init_water($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_water!\n" unless (defined($name));
    CORE::die "No information found for water '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('water_'.$name));
    $self->register_water($self->new_Water($name));
    return $self->water($name);
}

sub init_temp($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_temp!\n" unless (defined($name));
    CORE::die "No information found for temperature log '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('temp_'.$name));
    $self->register_temp($self->new_Templog($name));
    return $self->temp($name);
}

sub init_gasgroup($)
{
    my $self = shift;
    my $name = shift;
    CORE::die "No name defined for init_gasgroup!\n" unless (defined($name));
    CORE::die "No information found for gas group '$name'".' for rig '.$self->name."\n" unless ($self->{'_config'}->section_exists('gas_group_'.$name));
    $self->register_gasgroup($self->new_GasGroup($name));
    return $self->gasgroup($name);
}

sub init_device($$)
{
    my $self = shift;
    my $type = shift;
    my $name = shift;
    return undef unless ($type && $name);
    my $dev = $self->get_device($type,$name);
    return $dev if ($dev);
    ### Find if requested device is of filter type which may masquerade as normal type ###
    foreach my $n ($self->avail_filter_type($type))
    {
	next unless ($n eq $name);
	$dev = $self->init_filter($name);
	last;
    }
    return $dev if ($dev);
    ### Allocate device normally ###
    return $self->init_simplechannel($name) if ($type eq 'simplechannel');
    return $self->init_relay($name) if ($type eq 'relay');
    return $self->init_temp($name) if ($type eq 'templog');
    return $self->init_tempcontrol($name) if ($type eq 'tempcontrol');
    return $self->init_MFC($name) if ($type eq 'MFC');
    return $self->init_water($name) if ($type eq 'water');
    return $self->init_gas($name) if ($type eq 'gas');
    return $self->init_gasgroup($name) if ($type eq 'gasgroup');
    return $self->init_PSU($name) if ($type eq 'PSU');
    return $self->init_analog($name) if ($type eq 'analog');
    return $self->init_multiplex($name) if ($type eq 'multiplex');
    return $self->init_filter($name) if ($type eq 'filter');
    return $self->init_logic($name) if ($type eq 'logic');
    return $self->init_PID($name) if ($type eq 'PID');
    return $self->init_Alert($name) if ($type eq 'Alert');
    return undef;
}

#### The new_Device functions are wrapper functions for the various Device.pm factory functions ###
#### They are used by $rig->init() as well as to obtain an device node for a not enabled device ###
#### (For testing purposes when installing a ned hwardware/logical device)                      ###

sub new_AnalogOut($)
{
    my $self = shift;
    my $name = shift;
    my $type = $self->get_cv('analog_out_'.$name,'type');
    return RFC::Device::new_AnalogOut($self,$name,$type);
}

sub new_PID($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_PID($self,$name);
}

sub new_Alert($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_Alert($self,$name);
}

sub new_filter($)
{
    my $self = shift;
    my $name = shift;
    my $type = $self->get_cv('filter_'.$name,'class');
    $type = 'input_spline' unless ($type);
    return RFC::Device::new_filter($self,$name,$type);
}

sub new_Logic($)
{
    my $self = shift;
    my $name = shift;
    my $type = $self->get_cv('Logic_'.$name,'class');
    $type = 'AND' unless ($type);
    return RFC::Device::new_Logic($self,$name,$type);
}

sub new_RELAY($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_RELAY($self,$name);
}

sub new_SimpleChannel($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_SimpleChannel($self,$name);
}

sub new_PSU($)
{
    my $self = shift;
    my $name = shift;
    my $type = $self->get_cv('PSU_'.$name,'type');
    ### Assume default Delta SM15-100 ###
    $type = 'SM_15_100' unless ($type);
    $type = shift if (@_);
    return RFC::Device::new_PSU($self,$name,$type);
}

sub new_Multiplexer($)
{
    my $self = shift;
    my $name = shift;
    my $type = $self->get_cv('gas_multiplexer_'.$self->name,'type');
    $type = 'ICP' unless ($type);
    return RFC::Device::new_Multiplexer($self,$name,$type);
}

sub new_MFC($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_MFC($self,$name);
}

sub new_TEMP($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_TEMP($self,$name);
}

sub new_GAS($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_GAS($self,$name);
}

sub new_Water($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_Water($self,$name);
}

sub new_Templog($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_Templog($self,$name);
}

sub new_GasGroup($)
{
    my $self = shift;
    my $name = shift;
    return RFC::Device::new_GasGroup($self,$name);
}


#### End Device.pm wrapper functions ###

sub set_temp($)
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $temp = shift;
    eval {
	$self->die("No temperature controler found!") unless ($self->tempcontrols);
	my $name;
	if (@_)
	{
	    $name = shift;
	    $self->die("Temperature controler '$name' not found!") unless exists($self->{'_TEMPCONTROL'}{$name});
	}
	else
	{	
	    $self->die("Ambigious temperature controler for setting temp, more than one controler found but no name specified!\n") if ($self->tempcontrols > 1);
	    my @list = $self->tempcontrols;
	    $name = $list[0];
	}
	$self->{'_TEMPCONTROL'}{$name}->set_temp($temp,'dummy','nolog');
	$self->proglog("Temperature for controler $name set to $temp C",@_);
    };
    if ($@)
    {
	$self->proglog("Error setting temperature: $@",@_);
    }
}

sub set_ramp($)
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $ramp = shift;
    eval {
	$self->die("No temperatue controler found!") unless ($self->tempcontrols);
	my $name;
	if (@_)
	{
	    $name = shift;
	    $self->die("Temperature controler '$name' not found!") unless exists($self->{'_TEMPCONTROL'}{$name});
	}
	else
	{	
	    $self->die("Ambigious temperature controler for setting temp, more than one controler found but no name specified!\n") if ($self->tempcontrols > 1);
	    my @list = $self->tempcontrols;
	    $name = $list[0];
	}
	$self->{'_TEMPCONTROL'}{$name}->set_ramp($ramp,'dummy','nolog');
	$self->proglog("Temperature ramp for controler '$name' set to $ramp C/hour",@_);
    };
    if ($@)
    {
	$self->proglog("Error setting ramprate: $@",@_);
    }
}

sub current($)
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $current = shift;
    eval {
	$self->die("No PSU found!") unless ($self->PSUs);
	my $name;
	if (@_)
	{
	    $name = shift;
	    $self->die("PSU '$name' not found!") unless exists($self->{'_PSU'}{$name});
	}
	else
	{
	    $self->die("Ambigious PSU for setting current, more than one PSU found but no name specified!\n") if ($self->PSUs > 1);
	    my @list = $self->PSUs;
	    $name = $list[0];
	}
	$self->{'_PSU'}{$name}->current($current,@_);
    };
    if ($@)
    {
	$self->proglog("Error setting current: $@",@_);
    }    
}

sub voltage($)
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $volt = shift;
    eval {
	$self->die("No PSU found!") unless ($self->PSUs);
	my $name;
	if (@_)
	{
	    $name = shift;
	    $self->die("PSU '$name' not found!") unless exists($self->{'_PSU'}{$name});
	}
	else
	{
	    $self->die("Ambigious PSU for setting current, more than one PSU found but no name specified!\n") if ($self->PSUs > 1);
	    my @list = $self->PSUs;
	    $name = $list[0];
	}
	$self->{'_PSU'}{$name}->voltage($volt,@_);
    };
    if ($@)
    {
	$self->proglog("Error setting voltage: $@",@_);
    }    
}


sub get_sessionlog()
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $session = shift;
    $session = $self->session unless ($session);
    $self->die("Test '$session' does not exist on rig ".$self->name."! (valid values are 0 to ".$self->session.')') if ($session < 0 || $session > $self->session);
    if ($session == $self->session)
    {
        return join("\n",$self->{'_userlog'}->readlines) if ($self->{'_userlog'}->exist);
    }
    else
    {
	my $webtest = $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$session.'/';
        my $file = SemaforeFile->new($webtest.'proglog',$self->{'_riglock'});
	my $l = $file->lock_sh_nb;
	unless ($l)
	{
	    sleep(1);
	    $l = $file->lock_sh_nb;
	}
	if ($l)
	{
	    my $str = join("\n",$file->readlines) if ($file->exist);
	    $file->lock_un;
	    return $str;
	}
	else
	{
	    warn "Could not lock session log file for reading!\n";
	    return '';
	}
	$file->DESTROY();
    }
    return '';
}

sub get_errorlog()
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $session = shift;
    $session = $self->session unless ($session);
    $self->die("Test '$session' does not exist on rig ".$self->name."! (valid values are 0 to ".$self->session.')') if ($session < 0 || $session > $self->session);
    if ($session == $self->session)
    {
	my $l = $self->{'_errlog'}->lock_sh_nb;
	unless ($l)
	{
	    sleep(1);
	    $l = $self->{'_errlog'}->lock_sh_nb;
	}
	if ($l)
	{
	    my $str = join("\n",$self->{'_errlog'}->readlines) if ($self->{'_errlog'}->exist);
	    $self->{'_errlog'}->lock_un;
	    return $str;
	}
	else
	{
	    warn "Could not lock errorlog file for reading!\n";
	    return '';	    
	}
    }
    else
    {
	my $webtest = $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$session.'/';
	my $file = SemaforeFile->new($webtest.'errorlog',$self->{'_riglock'});
	my $l = $file->lock_sh_nb;
	unless ($l)
	{
	    sleep(1);
	    $l = $file->lock_sh_nb;
	}
	if ($l)
	{
	    my $str = join("\n",$file->readlines) if ($file->exist);
	    $file->lock_un;
	    return $str;
	}
	else
	{
	    warn "Could not lock errorlog file for session $session for reading!\n";
	    return '';
	}
    }
    return '';
}

sub config_key_exists($$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    return $self->{'_config'}->exists($section,$key);
}

sub config_section_exists($$)
{
    my $self = shift;
    my $section = shift;
    return $self->{'_config'}->section_exists($section);
}

sub change_config_value($$$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    my $newvalue = shift;
    return $self->{'_config'}->change_config_value($section,$key,$newvalue);
}

sub get_cv($$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    return $self->{'_config'}->get_config_value($section,$key);
}

sub get_cv_list($$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    my $str = $self->get_cv($section,$key);
    return () unless ($str);
    return split(/\s*\,\s*/,$str);
}

sub get_cv_test($$$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    my $session = shift;
    $session = $self->session unless (defined($session));
    $self->die("Test '$session' does not exist on rig ".$self->name."! (valid values are 0 to ".$self->session.')') if ($session < 0 || $session > $self->session);
    unless ($self->{'_config_old'}[$session] && ref($self->{'_config_old'}[$session]))
    {
	my $fn = $self->{'_config_old'}[$session];
	$self->{'_config_old'}[$session] = ElchemeaConfig->new($fn,$self->{'_configlock'});
	$self->{'_config_old'}[$session]->debug($self->{'_debug'}) if ($self->{'_config_old'}[$session] && ref($self->{'_config_old'}[$session]));
    }
    return $self->{'_config_old'}[$session]->get_config_value($section,$key);
}

sub name()
{
    my $self = shift;
    return $self->{'_name'};
}

sub session()
{
    my $self = shift;
    return $self->{'_session'};
}

sub errorlog
{
    my $self = shift;
    $self->proglog(@_);
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $string = shift;
    chomp $string;
    my $do_trace = 0;
    my $s;
    my $l = $self->{'_errlog'}->lock_ex_nb();
    unless ($l)
    {
	sleep(1);
	$l = $self->{'_errlog'}->lock_ex_nb();
    }
    if ($l)
    {
	unless ($self->{'_errlog'}->exist)
	{
	    $self->writedebug("Writing to new errorlog for test ".$self->session." for rig '".$self->name."'") if ($self->debug);
	    $self->{'_errlog'}->writeline();
	    chmod 0666,$self->{'_errlog'}->filename;
	}
	if (@_)
	{
	    my $user = shift;
	    ### Assume second argument is user name and append ' by user' ###
	    $string .= " by user '$user'";
	}
	else
	{
	    $do_trace = 1;
	    my $tty = POSIX::ttyname(1);
	    my $user = RFC::Main::get_user_name($<);
	    foreach ($tty,$user)
	    {
		$_ = '' unless (defined($_));
	    }
	    $string .= " by user '$user' on $tty";
	}
	### Opens the logfile #####
	my $time = localtime;
	$s = $time.' : '.$string;
	$self->{'_errlog'}->append($s);
	$self->{'_errlog'}->unlock;
	$self->tracelog($string) if ($do_trace || $self->debug);
	return 1;
    }
    else
    {
	warn "Could not lock errorlog file for writing! Message '$s' Could not be appended!\n";
	return 0;	
    }
}

sub proglog
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my $string = shift;
    chomp $string;
    my $do_trace = 0;
    my $s;
    my $l = $self->{'_userlog'}->lock_ex_nb();
    unless ($l)
    {
	sleep(1);
	$l = $self->{'_userlog'}->lock_ex_nb();
    }
    if ($l)
    {
	unless ($self->{'_userlog'}->exist)
	{
	    $self->writedebug("Writing to new userlog for test ".$self->session." for rig '".$self->name."'") if ($self->debug);
	    $self->{'_userlog'}->writeline();
	    chmod 0666,$self->{'_userlog'}->filename;
	}
	if (@_)
	{
	    my $user = shift;
	    ### Assume second argument is user name and append ' by user' ###
	    $string .= " by user '$user'";
	}
	else
	{
	    $do_trace = 1;
	    my $tty = POSIX::ttyname(1);
	    my $user = RFC::Main::get_user_name($<);
	    foreach ($tty,$user)
	    {
		$_ = '' unless (defined($_));
	    }
	    $string .= " by user '$user' on $tty";
	}
	### Opens the logfile #####
	my $time = localtime;
	my $s = $time.' : '.$string;
	$self->{'_userlog'}->append($s);
	$self->{'_userlog'}->unlock;
	$self->tracelog($string) if ($do_trace || $self->debug);
	return 1;
    }
    else
    {
	warn "Could not lock userlog file for writing! Message '$s' Could not be appended!\n";
	return 0;	
    }
}

sub tracelog
{
    my $self = shift;
    ### NB Do NOT call init here as this may leed to deep recursion! ###
    my $string = shift;
    chomp $string;
    $string = '' unless defined($string);
    #######################################
    ## Getting complete stack backtrace ###
    #######################################
    my @trace = ();
    my $i = 0;
    while(my ($package,$filename,$line,$subroutine) = caller($i++))
    {
        unshift @trace,"'$filename' line $line called $subroutine";
    }
    my $stacktrace = join(' -> ',@trace);
    $string .= ' #### Stack backtrace: '.$stacktrace;
    ### Opens the logfile #####
    my $time = localtime;
    my $s = $time.' : '.$string;
    my $l = $self->{'_trace'}->lock_ex_nb();
    unless ($l)
    {
	sleep(1);
	$l = $self->{'_trace'}->lock_ex_nb();
    }
    if ($l)
    {
	$self->{'_trace'}->lock_un();
	unless ($self->{'_trace'}->exist)
	{
	    warn "Writing to new tracelog for test ".$self->session." for rig '".$self->name."'\n" if ($self->debug);
	    $self->{'_trace'}->writeline();
	    chmod 0666,$self->{'_trace'}->filename;
	}
	$self->{'_trace'}->append($s);
	$self->{'_trace'}->unlock;
	return 1;
    }
    else
    {
	warn "Could not lock tracelog file for writing! Message '$s' Could not be appended!\n";
	return 0;
    }
}


sub load_session_nr()
### Function returning the session number ###
{
    my $self = shift;
    my $file = $publicdir.'rig'.$self->{'_name'}.'/test.nr';
    $self->{'_session'} = 0;
    if (-e $file)
    {
        $self->writedebug("Loading test number from file '$file'") if ($self->debug);
        sysopen(FH, $self->{'_riglock'}, O_WRONLY | O_CREAT) or $self->die("Could not open lockfile $self->{'_riglock'} $!");
        my $l = flock(FH, LOCK_SH|LOCK_NB);
	unless ($l)
	{
	    sleep(1);
	    $l = flock(FH, LOCK_SH|LOCK_NB);
	}
	if ($l)
	{
	    open MYIN,'<',$file or $self->die("Could not open file $file $!");
	    my $test = <MYIN>;
	    CORE::close MYIN;
	    flock(FH, LOCK_UN);
	    CORE::close FH;
	    if ($test =~ /(\d+)/)
	    {
		$self->{'_session'} = $1;
	    }
	}
	else
	{
	    CORE::close FH;
	    my $fn = $self->{'_riglock'};
	    warn "Could not lock '$fn' for reading when loading session number!\n";
	    return 0;
	}
    }
    $self->{'_webtest'} = $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$self->session.'/';
    return $self->{'_session'};
}


sub print_setup()
{
    #############################################
    ### print_setup()                         ###
    ### utility function for debugging        ###
    ### purposes, prints the contents of the  ###
    ### device instance data fields $self->{} ###
    #############################################
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my @arr = ('*******************************************');
    push @arr,'Rig ID: Rig'.$self->{'_name'};
    push @arr,'Member functions:';
    my $class = $self;
    $class =~ s/=.*//;
    my @methods = Class::Inspector->methods( $class, 'full', 'public' );
    foreach my $m (@methods)
    {
        if (ref($m) =~ /ARRAY/)
        {
            foreach (@{$m})
            {
                push @arr,"\t$_";
            }
            next;
        }
        push @arr,"\t$m";
    }
    push @arr,'';
    push @arr,'Instance data (Field,  Value):';
    foreach my $key (keys %{$self})
    {
        if (defined ($self->{$key}))
        {
            if (ref($self->{$key}) =~ /ARRAY/)
            {
                push @arr,"$key, Array with values:";
                push @arr,"\tIndex\tValue";
                my $x = 0;
                foreach (@{$self->{$key}})
                {
                    push @arr,"\t$x\t'$_'";
                    $x++;
                }
                push @arr,'';
                next;
            }
            if (ref($self->{$key}) =~ /HASH/)
            {
                push @arr,"$key, Asociative array with values:";
                push @arr,"\tKey\tValue";
                foreach my $k (keys %{$self->{$key}})
                {
                    push @arr,"\t'$k'\t'$self->{$key}{$k}'";
                }
                push @arr,'';
                next;
            }
            push @arr,$key.",\t'$self->{$key}'";
        }
        else
        {
            push @arr,$key.",\t\\N";
        }
    }
    push @arr,'';
    return join("\n",@arr);
}

sub test_list()
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    $self->die('Intialisation of rig '.$self->name.' failed, last mesage: '.$self->{'_error'}) if ($self->{'_error'});
    my @arr = ();
    foreach my $key ('_MULTI','_MFC','_RELAY','_SIMPLE','_WATER','_GAS','_TEMPLOG','_GASGROUP','_PSU','_AOUT','_TEMPCONTROL','_FILTER','_PID','_LOGIC','_ALERT')
    {
	foreach my $k (keys %{$self->{$key}})
	{
	    my %item = ('id'=>$self->{$key}{$k},'name'=>$key.'_'.$self->{$key}{$k}->name);
	    push @arr,\%item;
	}
    }
    return @arr;
}


sub print_config()
{
    my $self = shift;
    $self->init() unless ($self->{'_modtime'} == $self->modtime());
    my @arr = ('*******************************************');
    push @arr,'Rig ID: Rig'.$self->{'_name'};
    ### Ignore temperatue as it is only control and the Eutotherm module does not support print_config() ###
    foreach my $key ('_MULTI','_MFC','_RELAY','_SIMPLE','_WATER','_GAS','_TEMPLOG','_GASGROUP','_PSU','_AOUT','_TEMPCONTROL','_FILTER','_PID','_LOGIC','_ALERT')
    {
	push @arr , '****** '.$key.' ******';
	foreach my $k (keys %{$self->{$key}})
	{
	    push @arr,$self->{$key}{$k}->print_config;
	}
    }
    push @arr,'';
    return join("\n",@arr);
}

sub check_run()
{
    my $self = shift;
    my $usernumber = $<;
    my $username = RFC::Main::get_user_name($usernumber);
    $self->die('You are attempting to access processes for rig '.$self->name.', but are not logged in as rig'.$self->name) unless ($username eq 'rig'.$self->name);
    return 1;
}


sub inc_test($)
{
    my $self = shift;
    my $text = shift;
    my $user = shift;
    $self->lock_control;
    my $session = $self->session();
    my $newsession = $session + 1;
    #### Increase test nmber ###
    my $file = $publicdir.'rig'.$self->{'_name'}.'/test.nr';
    $self->writedebug("Writing new test number to file '$file'") if ($self->debug);
    open MYOUT,'>',$file or $self->die("Could not open file $file  for writing$!");
    print MYOUT $newsession,"\n";
    close MYOUT;
    $self->load_session_nr;
    $self->create_dirs;
    ### Copy curret configuration to new session config ###
    ### Wil be overwritten once a day                   ###
    my $fn = $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$newsession.'/configfile';
    $self->{'_config_old'}[$newsession] = ElchemeaConfig->new($fn,$self->{'_configlock'});
    $self->{'_config_old'}[$newsession]->debug($self->{'_debug'});
    my @lines = $self->config->readlines();
    $self->{'_config_old'}[$newsession]->writeline(@lines);
    chmod 0664,$self->{'_config_old'}[$newsession]->filename;
    $self->init();
    $self->create_dirs;
    my $string = "New test started";
    my $time = localtime();
    my $r = $self->name;
    $self->proglog($string,$user);
    $string = <<END;
Information for test $newsession on rig $r

Test started $time

Rignumber                   : $r
Testnumber                  : $newsession
$text

END

    $self->info($string);
    foreach my $k (keys %{$self->{'_GAS'}})
    {
	$self->{'_GAS'}{$k}->setflow(0,$user);
    }
    foreach ($self->PSUs)
    {
	$self->current('ocv',$_,$user);
    }
    my $olddir = $RFC::Header::publicdir.'rig'.$self->name.'/'.$self->name.'test'.$session.'/';
    my $target = $homedir.'infofiles/'.$self->name.'test'.$session;
    if (-e $olddir.'proglog')
    {
	copy($olddir.'proglog',$target.'proglog') or $self->die("Could not copy file $olddir"."proglog $!");
    }
    if (-e $olddir.'info.txt')
    {
	copy($olddir.'info.txt',$target.'.inf') or $self->die("Could not copy file $olddir"."info.txt $!");
    }
    ### Do one measurement in new test ###
    RFC::Main::CGI_client($self->name,'measure');
    ### Set mail lock files for vogt and H2vogt ###
    $self->set_warning('H2mailfile');
    $self->set_warning('mailfile');
    ### Mail users that new test has been started ###
    my @adrs = split(/\,/,RFC::Main::get_cv('celltest','start_test_mail_users'));
    foreach (@adrs)
    {
	my $mok = RFC::Main::mail($_,"New test started on rig $r\n Current test: $r test $newsession\n$text","New test started on rig $r");
	$self->proglog("Warning: Start test mail not sent to '$_'") unless ($mok);
    }
    $self->unlock_control;
    return $self->session;
}



sub create_dirs()
{
    my $self = shift;
    my $session = $self->session();
    my $webdir = $publicdir.'rig'.$self->name.'/'.$self->name.'test'.$self->session.'/';
    unless (-d $webdir)
    {
	mkdir $webdir,0775;
	chmod 0775, $webdir;
    }
    my $impdir = $webdir .'impedance/';
    unless (-d $impdir)
    {
	mkdir $impdir,0775;
	chmod 0775, $impdir;
    }
    return;
}
